# coding=utf-8
"""Volume Flow Rate data type."""
from __future__ import division

from .base import DataTypeBase
from .volume import Volume
from .volumeflowrateintensity import VolumeFlowRateIntensity


class VolumeFlowRate(DataTypeBase):
    """Volume Flow Rate
    """
    _units = ('m3/s', 'ft3/s', 'L/s', 'cfm', 'gpm', 'mL/s', 'fl oz/s', 'L/h', 'gph')
    _si_units = ('m3/s', 'L/s', 'mL/s', 'L/h')
    _ip_units = ('ft3/s', 'cfm', 'gpm', 'fl oz/s', 'gph')
    _min = 0
    _abbreviation = 'dV/dt'
    _normalized_type = VolumeFlowRateIntensity

    def _m3_s_to_ft3_s(self, value):
        return value * 35.3147

    def _m3_s_to_L_s(self, value):
        return value * 1000.

    def _m3_s_to_cfm(self, value):
        return value * 2118.88

    def _m3_s_to_gpm(self, value):
        return value * 15850.3231

    def _m3_s_to_mL_s(self, value):
        return value * 1000000.

    def _m3_s_to_floz_s(self, value):
        return value * 33814.

    def _m3_s_to_L_h(self, value):
        return value * 3600000.

    def _m3_s_to_gph(self, value):
        return value * 951019.

    def _ft3_s_to_m3_s(self, value):
        return value / 35.3147

    def _L_s_to_m3_s(self, value):
        return value / 1000.

    def _cfm_to_m3_s(self, value):
        return value / 2118.88

    def _gpm_to_m3_s(self, value):
        return value / 15850.3231

    def _mL_s_to_m3_s(self, value):
        return value / 1000000.

    def _floz_s_to_m3_s(self, value):
        return value / 33814.

    def _L_h_to_m3_s(self, value):
        return value / 3600000.

    def _gph_to_m3_s(self, value):
        return value / 951019.

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('m3/s', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'L/s':
            return self.to_unit(values, 'cfm', from_unit), 'cfm'
        elif from_unit == 'mL/s':
            return self.to_unit(values, 'fl oz/s', from_unit), 'fl oz/s'
        elif from_unit == 'L/h':
            return self.to_unit(values, 'gph', from_unit), 'gph'
        else:
            return self.to_unit(values, 'ft3/s', from_unit), 'ft3/s'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'cfm':
            return self.to_unit(values, 'L/s', from_unit), 'L/s'
        elif from_unit == 'fl oz/s':
            return self.to_unit(values, 'mL/s', from_unit), 'mL/s'
        elif from_unit == 'gph':
            return self.to_unit(values, 'L/h', from_unit), 'L/h'
        else:
            return self.to_unit(values, 'm3/s', from_unit), 'm3/s'



# coding=utf-8
"""Time data type."""
from __future__ import division

from .base import DataTypeBase


class Time(DataTypeBase):
    """Time"""
    _units = ('hr', 'min', 'sec', 'day')
    _si_units = ('hr', 'min', 'sec', 'day')
    _ip_units = ('hr', 'min', 'sec', 'day')
    _cumulative = True
    _abbreviation = 't'

    def _hr_to_min(self, value):
        return value * 60.

    def _hr_to_sec(self, value):
        return value * 3600.

    def _hr_to_day(self, value):
        return value / 24.

    def _min_to_hr(self, value):
        return value / 60.

    def _sec_to_hr(self, value):
        return value / 3600.

    def _day_to_hr(self, value):
        return value * 24.

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('hr', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        return values, from_unit

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        return values, from_unit



# coding=utf-8
"""Illuminance data type."""
from __future__ import division

from .base import DataTypeBase


class Illuminance(DataTypeBase):
    """Illuminance
    """
    _units = ('lux', 'fc')
    _si_units = ('lux')
    _ip_units = ('fc')
    _min = 0
    _abbreviation = 'Ev'
    _point_in_time = False

    def _lux_to_fc(self, value):
        return value / 10.7639

    def _fc_to_lux(self, value):
        return value * 10.7639

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('lux', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit == 'fc':
            return values, from_unit
        else:
            return self.to_unit(values, 'fc', from_unit), 'fc'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit == 'lux':
            return values, from_unit
        else:
            return self.to_unit(values, 'lux', from_unit), 'lux'


class GlobalHorizontalIlluminance(Illuminance):
    _abbreviation = 'GHI'


class DirectNormalIlluminance(Illuminance):
    _abbreviation = 'DNI'


class DiffuseHorizontalIlluminance(Illuminance):
    _abbreviation = 'DHI'



# coding=utf-8
"""Module for computing geometry for the compass used by a variety of graphics."""
from __future__ import division

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.arc import Arc2D
from ladybug_geometry.geometry3d.pointvector import Point3D

import math


class Compass(object):
    """Object for computing geometry for the compass used by a variety of graphics.

    Methods to project points to orthographic and stereographic projections are
    also within this class so that "domed" visualizations can be synchronized with
    the compass in the 2D plane.

    Args:
        radius: A positive number for the radius of the compass. (Default: 100).
        center: A ladybug_geometry Point2D for the center of the compass
            in the scene. (Default: (0, 0) for the World origin).
        north_angle: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East (Default: 0).
        spacing_factor: A positive number for the fraction of the radius that
            labels and tick marks occupy around the compass. (Default: 0.15)

    Properties:
        * radius
        * center
        * north_angle
        * north_vector
        * spacing_factor
        * min_point
        * max_point
        * inner_boundary_circle
        * all_boundary_circles
        * major_azimuth_points
        * major_azimuth_ticks
        * minor_azimuth_points
        * minor_azimuth_ticks
        * orthographic_altitude_circles
        * orthographic_altitude_points
        * stereographic_altitude_circles
        * stereographic_altitude_points
    """
    __slots__ = ('_radius', '_center', '_north_angle', '_north_vector',
                 '_spacing_factor')

    MAJOR_AZIMUTHS = (0, 90, 180, 270)
    MAJOR_TEXT = ('N', 'E', 'S', 'W')
    MINOR_AZIMUTHS = (22.5, 45, 67.5, 112.5, 135, 157.5, 202.5, 225, 247.5,
                      292.5, 315, 337.5)
    MINOR_TEXT = ('NNE', 'NE', 'ENE', 'ESE', 'SE', 'SSE', 'SSW', 'SW', 'WSW',
                  'WNW', 'NW', 'NNW')
    ALTITUDES = (10, 20, 30, 40, 50, 60, 70, 80)
    PI = math.pi

    def __init__(self, radius=100, center=Point2D(), north_angle=0,
                 spacing_factor=0.15):
        """Initialize Compass."""
        self.radius = radius
        self.center = center
        self.north_angle = north_angle
        self.spacing_factor = spacing_factor

    @property
    def radius(self):
        """Get or set a positive number for the radius of the compass."""
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = float(value)
        assert self._radius > 0, \
            'Compass radius must be a positive number. Got {}.'.format(value)

    @property
    def center(self):
        """Get or set a Point2D for the center of the compass in the scene."""
        return self._center

    @center.setter
    def center(self, value):
        assert isinstance(value, Point2D), 'Expected ladybug_geometry Point2D ' \
            'for Compass center. Got {}.'.format(type(value))
        self._center = value

    @property
    def north_angle(self):
        """Get or set a number between -360 and 360 for the north_angle in degrees."""
        return math.degrees(self._north_angle)

    @north_angle.setter
    def north_angle(self, value):
        self._north_angle = math.radians(float(value))
        self._north_vector = Vector2D(0, 1).rotate(math.radians(-self._north_angle))
        assert -self.PI * 2 <= self._north_angle <= self.PI * 2, \
            'north_angle value should be between -360 and 360. Got {}.'.format(value)

    @property
    def north_vector(self):
        """Get or set a ladybug_geometry Vector2D for the north direction."""
        return self._north_vector

    @north_vector.setter
    def north_vector(self, value):
        assert isinstance(value, Vector2D), \
            'Expected Vector2D for north_vector. Got {}.'.format(type(value))
        self._north_vector = value
        self._north_angle = \
            math.degrees(Vector2D(0, 1).angle_clockwise(self._north_vector))

    @property
    def spacing_factor(self):
        """Get or set a number for the fraction of radius occupied by labels and ticks.
        """
        return self._spacing_factor

    @spacing_factor.setter
    def spacing_factor(self, value):
        self._spacing_factor = float(value)
        assert self._spacing_factor > 0, \
            'Compass spacing_factor must be a positive number. Got {}.'.format(value)

    @property
    def min_point(self):
        """Get a Point2D for the minimum around the entire compass."""
        fac = (1 + self.spacing_factor) * self.radius
        return Point2D(self.center.x - fac, self.center.y - fac)

    @property
    def max_point(self):
        """Get a Point2D for the minimum around the entire compass."""
        fac = (1 + self.spacing_factor) * self.radius
        return Point2D(self.center.x + fac, self.center.y + fac)

    @property
    def inner_boundary_circle(self):
        """Get a Arc2D for the inner circle of the compass.

        This is essentially a circle with the compass radius.
        """
        return Arc2D(self.center, self.radius)

    @property
    def all_boundary_circles(self):
        """Get an array of 3 Arc2Ds for the circles of the compass."""
        arc2 = Arc2D(self.center, self.radius * (1 + self.spacing_factor * 0.1))
        arc3 = Arc2D(self.center, self.radius * (1 + self.spacing_factor * 0.3))
        return [self.inner_boundary_circle, arc2, arc3]

    @property
    def major_azimuth_points(self):
        """Get a list of Point2Ds for the major azimuth labels."""
        return self.label_points_from_angles(self.MAJOR_AZIMUTHS)

    @property
    def major_azimuth_ticks(self):
        """Get a list of LineSegment2Ds for the major azimuth labels."""
        return self.ticks_from_angles(self.MAJOR_AZIMUTHS, 0.5)

    @property
    def minor_azimuth_points(self):
        """Get a list of Point2Ds for the minor azimuth labels."""
        return self.label_points_from_angles(self.MINOR_AZIMUTHS)

    @property
    def minor_azimuth_ticks(self):
        """Get a list of LineSegment2Ds for the minor azimuth labels."""
        return self.ticks_from_angles(self.MINOR_AZIMUTHS, 0.3)

    @property
    def orthographic_altitude_circles(self):
        """Get a list of circles for the orthographic altitude labels."""
        circles = []
        for angle in self.ALTITUDES:
            circles.append(
                Arc2D(self.center, self.radius * math.cos(math.radians(angle)))
            )
        return circles

    @property
    def orthographic_altitude_points(self):
        """Get a list of Point2Ds for the orthographic altitude labels."""
        pts = []
        for angle in self.ALTITUDES:
            spacing_fac = self.radius * 0.01  # spacing factor
            add_y = (self.radius * math.cos(math.radians(angle))) - spacing_fac
            pts.append(Point2D(self.center.x, self.center.y + add_y))
        if self._north_angle != 0:
            pts = [pt.rotate(self._north_angle, self.center) for pt in pts]
        return pts

    @property
    def stereographic_altitude_circles(self):
        """Get a list of circles for the stereographic altitude labels."""
        circles = []
        for angle in self.ALTITUDES:
            ang = math.radians(angle)
            pt3d = Point3D(math.cos(ang), 0, math.sin(ang))
            radius = self.point3d_to_stereographic(pt3d, 1).x * self.radius
            circles.append(Arc2D(self.center, radius))
        return circles

    @property
    def stereographic_altitude_points(self):
        """Get a list of Point2Ds for the stereographic altitude labels."""
        pts = []
        for angle in self.ALTITUDES:
            spacing_fac = self.radius * 0.01  # spacing factor
            ang = math.radians(angle)
            pt3d = Point3D(math.cos(ang), 0, math.sin(ang))
            add_y = (self.point3d_to_stereographic(pt3d, 1).x * self.radius) \
                - spacing_fac
            pts.append(Point2D(self.center.x, self.center.y + add_y))
        if self._north_angle != 0:
            pts = [pt.rotate(self._north_angle, self.center) for pt in pts]
        return pts

    def label_points_from_angles(self, angles, factor=0.8):
        """Get a list of label points from a list of angles between 0 and 360.

        Args:
            angles: An array of numbers between 0 and 360 for the angles of
                custom angle labels to be generated for the compass.
            factor: A number between 0 and 1 for the fraction of the spacing_factor
                at which the points should be generated.
        """
        circle = Arc2D(self.center, self.radius * (1 + self.spacing_factor * factor))
        return [circle.point_at_angle(self._north_angle - math.radians(angle - 90))
                for angle in angles]

    def ticks_from_angles(self, angles, factor=0.3):
        """Get a list of Linesegment2Ds from a list of angles between 0 and 360."""
        pts_in = self.label_points_from_angles(angles, 0)
        pts_out = self.label_points_from_angles(angles, factor)
        return [LineSegment2D.from_end_points(pi, po) for pi, po in zip(pts_in, pts_out)]

    def min_point3d(self, z=0):
        """Get a Point3D for the minimum around the entire compass."""
        min_pt = self.min_point
        return Point3D(min_pt.x, min_pt.y, z)

    def max_point3d(self, z=0):
        """Get a Point3D for the minimum around the entire compass."""
        max_pt = self.max_point
        return Point3D(max_pt.x, max_pt.y, z)

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    @staticmethod
    def point3d_to_orthographic(point):
        """Get a Point2D for a given Point3D using a orthographic projection.

        Args:
            point: A ladybug_geometry Point3D to be projected into 2D space via
                stereographic projection.
        """
        return Point2D(point.x, point.y)

    @staticmethod
    def point3d_to_stereographic(point, radius=100, origin=Point3D()):
        """Get a Point2D for a given Point3D using a stereographic projection.

        Args:
            point: A ladybug_geometry Point3D to be projected into 2D space via
                stereographic projection.
            radius: A positive number for the radius of the sphere on which the
                point exists. (Default: 100).
            origin: An optional ladybug_geometry Point3D representing the origin
                of the coordinate system in which the projection is happening.
                (eg. the center of the compass).
        """
        # move the point to the world origin
        coords = ((point.x - origin.x), (point.y - origin.y), (point.z - origin.z))
        # perform the stereographic projection while scaling it to the unit sphere
        proj_pt = (coords[0] / (radius + coords[2]), coords[1] / (radius + coords[2]))
        # move the point back to its original location and scale
        return Point2D(proj_pt[0] * radius + origin.x, proj_pt[1] * radius + origin.y)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.radius, hash(self.center), self.north_angle, self.spacing_factor)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Compass) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __copy__(self):
        return Compass(self.radius, self.center, self.north_angle, self.spacing_factor)

    def __repr__(self):
        """Compass representation."""
        return "Compass (radius:{}, center:{})".format(self.radius, self.center)



# coding=utf-8
"""Module of Data Types (eg. Temperature, Area, etc.)

Possesses capabilities for unit conversions and range checks.
It also includes descriptions of the data types and the units.

Properties:
    TYPES: A tuple indicating all currently supported data types.

    BASETYPES: A tuple indicating all base types. Base types are the
    data types on which unit systems are defined.

    UNITS: A dictionary containing all currently supported units. The
    keys of this dictionary are the base type names (eg. 'Temperature').

    TYPESDICT: A dictionary containing pointers to the classes of each data type.
    The keys of this dictionary are the data type names.
"""

from .base import _DataTypeEnumeration

_data_types = _DataTypeEnumeration(import_modules=True)
TYPES = _data_types.types
BASETYPES = _data_types.base_types
UNITS = _data_types.units
TYPESDICT = _data_types.types_dict



"""Utilities for serializing Data Collections to and from files."""
import os
import json

from ladybug.dt import DateTime
try:  # check if we are in IronPython
    import cPickle as pickle
except ImportError:  # wea re in cPython
    import pickle

from .datacollection import BaseCollection, HourlyDiscontinuousCollection, \
    HourlyContinuousCollection, DailyCollection, MonthlyCollection, \
    MonthlyPerHourCollection
from .header import Header
from .analysisperiod import AnalysisPeriod
from .dt import Date


def collections_to_csv(data_collections, folder, file_name='data.csv'):
    """Write a series of aligned Data Collections into a CSV.

    These can be serialized back to the original objects using the
    collections_from_csv method.

    Args:
        data_collections: A list of aligned Data Collections which will be
            written to a CSV.
        folder: Folder to which the CSV will be written.
        file_name: File name for the CSV (Default: data.csv).

    Returns:
        The path to the CSV file to which the data_collections have been written.
    """
    # check that the collections are aligned with each other
    assert BaseCollection.are_collections_aligned(data_collections, False), \
        'Data collections must be aligned with each other in order to ' \
        'use collections_to_csv.'
    header_len = 2 + len(data_collections[0].header.metadata) \
        if BaseCollection.are_metadatas_aligned(data_collections, False) else 3
    csv_data = []

    # create the first column with the datetimes
    dt_column = [''] * (header_len - 2)
    dt_column.append(data_collections[0]._collection_type)
    dt_column.append(str(data_collections[0].header.analysis_period))
    dt_column.extend(data_collections[0].datetime_strings)
    csv_data.append(dt_column)

    # loop through the data collections and add columns for each
    meta_per_row = False if header_len == 3 else True
    for dat in data_collections:
        dat_column = dat.header.to_csv_strings(meta_per_row)
        dat_column.extend([str(v) for v in dat.values])
        csv_data.append(dat_column)

    # dump all of the data into the CSV file
    if not os.path.isdir(folder):
        os.makedirs(folder)
    if not file_name.lower().endswith('.csv'):
        file_name += '.csv'
    file_path = os.path.join(folder, file_name)
    with open(file_path, 'w') as outf:
        for row in zip(*csv_data):
            outf.write(','.join(row) + '\n')
    return file_path


def collections_from_csv(data_file):
    """Load a series of Data Collections from a JSON file.

    Args:
        data_file: Path to a .json file containing a list of data collections.

    Returns:
        A list of data collections loaded from the .json file.
    """
    # perform checks and set variables to help with re-serialization
    assert os.path.isfile(data_file), 'Failed to find %s' % data_file
    coll_types = {
        'HourlyContinuous': HourlyContinuousCollection,
        'HourlyDiscontinuous': HourlyDiscontinuousCollection,
        'Monthly': MonthlyCollection,
        'Daily': DailyCollection,
        'MonthlyPerHour': MonthlyPerHourCollection
    }

    # load all of the data from the file
    headers, values, datetimes, coll_class, aper = [], [], [], None, None
    with open(data_file) as inf:
        # first load all of the header information
        for row in inf:
            row_data = row.strip().split(',')
            headers.append(row_data[1:])
            if row_data[0] in coll_types:
                coll_class = coll_types[row_data[0]]
            elif coll_class is not None:
                aper = AnalysisPeriod.from_string(row_data[0])
                break
        # then, load all of the values and datetimes
        for row in inf:
            row_data = row.split(',')
            datetimes.append(row_data[0])
            values.append([float(v) for v in row_data[1:]])

    # reconstruct data collections from the loaded data
    heads = [Header.from_csv_strings(h, aper) for h in zip(*headers)]
    t_vals = zip(*values)
    if coll_class == HourlyContinuousCollection:
        data = [HourlyContinuousCollection(h, v) for h, v in zip(heads, t_vals)]
    elif coll_class == HourlyDiscontinuousCollection:
        dts = [DateTime.from_date_time_string(d) for d in datetimes]
        data = [HourlyDiscontinuousCollection(h, v, dts) for h, v in zip(heads, t_vals)]
    elif coll_class == MonthlyCollection:
        inv_map = {v: k for k, v in AnalysisPeriod.MONTHNAMES.items()}
        dts = [inv_map[d] for d in datetimes]
        data = [MonthlyCollection(h, v, dts) for h, v in zip(heads, t_vals)]
    elif coll_class == DailyCollection:
        dts = [Date.from_date_string(d, aper.is_leap_year) for d in datetimes]
        data = [DailyCollection(h, v, dts) for h, v in zip(heads, t_vals)]
    elif coll_class == MonthlyPerHourCollection:
        inv_map = {v: k for k, v in AnalysisPeriod.MONTHNAMES.items()}
        dt_strs = [d.split(' ') for d in datetimes]
        dts = [(inv_map[d[0]], int(d[1].split(':')[0]), int(d[1].split(':')[1]))
               for d in dt_strs]
        data = [MonthlyPerHourCollection(h, v, dts) for h, v in zip(heads, t_vals)]
    for d in data:
        d._validated_a_period = True
    return data


def collections_to_json(data_collections, folder, file_name='data.json', indent=None):
    """Write a series of Data Collections into a JSON.

    These can be serialized back to the original objects using the
    collections_from_json method.

    Args:
        data_collections: A list of aligned Data Collections which will be
            written to a JSON.
        folder: Folder to which the JSON will be written.
        file_name: File name for the JSON (Default: data.json).
        indent: A positive integer to set the indentation used in the resulting
            JSON file. (Default: None).

    Returns:
        The path to the JSON file to which the data_collections have been written.
    """
    # convert the data collections to an array of dictionaries
    dat_dict = [dat.to_dict() for dat in data_collections]
    # dump all of the data into the JSON file
    if not os.path.isdir(folder):
        os.makedirs(folder)
    if not file_name.lower().endswith('.json'):
        file_name += '.json'
    file_path = os.path.join(folder, file_name)
    with open(file_path, 'w') as fp:
        json.dump(dat_dict, fp, indent=indent)
    return file_path


def collections_from_json(data_file):
    """Load a series of Data Collections from a JSON file.

    Args:
        data_file: Path to a .json file containing a list of data collections.

    Returns:
        A list of data collections loaded from the .json file.
    """
    assert os.path.isfile(data_file), 'Failed to find %s' % data_file
    with open(data_file) as inf:
        data = json.load(inf)
    return [_dict_to_collection(d_dict) for d_dict in data]


def collections_to_pkl(data_collections, folder, file_name='data.pkl'):
    """Write a series of Data Collections into a PKL.

    These can be serialized back to the original objects using the
    collections_from_pkl method.

    Args:
        data_collections: A list of aligned Data Collections which will be
            written to a PKL.
        folder: Folder to which the PKL will be written.
        file_name: File name for the PKL (Default: data.pkl).
        indent: A positive integer to set the indentation used in the resulting
            PKL file. (Default: None).

    Returns:
        The path to the PKL file to which the data_collections have been written.
    """
    # convert the data collections to an array of dictionaries
    dat_dict = [dat.to_dict() for dat in data_collections]
    # dump all of the data into the PKL file
    if not os.path.isdir(folder):
        os.makedirs(folder)
    if not file_name.lower().endswith('.pkl'):
        file_name += '.pkl'
    file_path = os.path.join(folder, file_name)
    with open(file_path, 'wb') as fp:
        pickle.dump(dat_dict, fp)
    return file_path


def collections_from_pkl(data_file):
    """Load a series of Data Collections from a PKL file.

    Args:
        data_file: Path to a .pkl file containing a list of data collections.

    Returns:
        A list of data collections loaded from the .pkl file.
    """
    assert os.path.isfile(data_file), 'Failed to find %s' % data_file
    with open(data_file, 'rb') as inf:
        data = pickle.load(inf)
    return [_dict_to_collection(d_dict) for d_dict in data]


def _dict_to_collection(data_dict):
    """Load any data collection dictionary to an object."""
    if data_dict['type'] == 'HourlyContinuous':
        return HourlyContinuousCollection.from_dict(data_dict)
    elif data_dict['type'] == 'HourlyDiscontinuous':
        return HourlyDiscontinuousCollection.from_dict(data_dict)
    elif data_dict['type'] == 'Monthly':
        return MonthlyCollection.from_dict(data_dict)
    elif data_dict['type'] == 'Daily':
        return DailyCollection.from_dict(data_dict)
    elif data_dict['type'] == 'MonthlyPerHour':
        return MonthlyPerHourCollection.from_dict(data_dict)
    else:
        raise ValueError(
            'Unrecognized data collection type "{}".'.format(data_dict['type']))



# coding=utf-8
"""Temperature Delta data type."""
from __future__ import division

from .base import DataTypeBase
from .temperaturetime import TemperatureTime


class TemperatureDelta(DataTypeBase):
    """TemperatureDelta
    """
    _units = ('dC', 'dF', 'dK')
    _si_units = ('dC', 'dK')
    _ip_units = ('dF')
    _abbreviation = 'DeltaT'
    _time_aggregated_type = TemperatureTime
    _time_aggregated_factor = 1. / 24.

    def _dC_to_dF(self, value):
        return value * 9. / 5.

    def _dC_to_dK(self, value):
        return value

    def _dF_to_dC(self, value):
        return value * (5. / 9.)

    def _dK_to_dC(self, value):
        return value

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('dC', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit == 'dF':
            return values, from_unit
        else:
            return self.to_unit(values, 'dF', from_unit), 'dF'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit == 'dC' or from_unit == 'dK':
            return values, from_unit
        else:
            return self.to_unit(values, 'dC', from_unit), 'dC'


class AirTemperatureDelta(TemperatureDelta):
    _abbreviation = 'DeltaTair'


class RadiantTemperatureDelta(TemperatureDelta):
    _abbreviation = 'DeltaTrad'


class OperativeTemperatureDelta(TemperatureDelta):
    _abbreviation = 'DeltaTo'



# coding=utf-8
"""Module for computing solar envelopes."""
from __future__ import division
import math
try:
    from itertools import izip as zip  # python 2
except:
    pass

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.ray import Ray2D
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.face import Face3D
from ladybug_geometry.geometry3d.mesh import Mesh3D


class SolarEnvelope(object):
    """Calculate a Solar Envelope boundary for a given site.

    Args:
        geometry_mesh: A ladybug geometry Mesh3D for which a solar envelope boundary
            is computed.
        obstacle_faces: A list of horizontal planar ladybug geometry Face3D
            indicating the tops (in the case of solar collection) or bottoms
            (in the case of solar rights) of context geometries. Being above a
            solar collection boundary ensures these top surfaces don't block
            the sun vectors to ones position. Being below a solar rights
            boundary ensures these bottom surfaces are protected from shade.
        sun_vectors: A list of ladybug geometry Vector3D for the sun vectors, which
            determine the hours of the year when sun should be accessible. These
            can be obtained from the ladybug.sunpath module.
        height_limit: A positive number for the minimum distance below (for collections)
            or maximum distance above (for rights) the average geometry_mesh height
            that the envelope points can be. This is used when there are no
            vectors blocked for a given point. (Default: 100).
        solar_rights: Set to True to compute a solar rights boundary and False to compute
            a solar collection boundary. Solar rights boundaries represent the
            boundary below which one can build without shading the surrounding
            obstacles from any of the sun_vectors. Solar collection boundaries
            represent the boundary above which the one will have direct solar
            access to all of the input sun_vectors. (Default: False).

    Properties:
        * geometry_mesh
        * obstacle_faces
        * sun_vectors
        * height_limit
        * max_height
        * solar_rights
        * geometry_point2ds
        * obstacle_polygon2ds
        * sun_vector2ds
        * sun_vector2ds_reversed
    """
    __slots__ = ('_geometry_mesh', '_obstacle_faces', '_sun_vectors',
                 '_height_limit', '_solar_rights')

    def __init__(self, geometry_mesh, obstacle_faces, sun_vectors,
                 height_limit=100, solar_rights=False):
        self.geometry_mesh = geometry_mesh
        self.obstacle_faces = obstacle_faces
        self.sun_vectors = sun_vectors
        self.height_limit = height_limit
        self.solar_rights = solar_rights

    @property
    def geometry_mesh(self):
        """Get or set a Mesh3D for the site for which a solar envelope is computed."""
        return self._geometry_mesh

    @geometry_mesh.setter
    def geometry_mesh(self, value):
        assert isinstance(value, Mesh3D), 'Expected ladybug_geometry ' \
            'Mesh3D. Got {}.'.format(type(value))
        self._geometry_mesh = value

    @property
    def obstacle_faces(self):
        """Get or set a tuple of Face3D for the context obstacles."""
        return self._obstacle_faces

    @obstacle_faces.setter
    def obstacle_faces(self, value):
        if not isinstance(value, tuple):
            value = tuple(value)
        assert len(value) > 0, 'There must be at least 1 obstacle face.'
        for face in value:
            assert isinstance(face, Face3D), \
                'Expected Face3D for obstacle_faces. Got {}.'.format(type(face))
        self._obstacle_faces = value

    @property
    def sun_vectors(self):
        """Get or set a tuple of Vector3D for the sun vectors for solar access."""
        return self._sun_vectors

    @sun_vectors.setter
    def sun_vectors(self, value):
        if not isinstance(value, tuple):
            value = tuple(value)
        assert len(value) > 0, 'There must be at least 1 sun vector.'
        for vec in value:
            assert isinstance(vec, Vector3D), \
                'Expected Vector3D for sun_vector. Got {}.'.format(type(vec))
        self._sun_vectors = value

    @property
    def height_limit(self):
        """Get or set a number for the height limit below or above the geometry_mesh."""
        return self._height_limit

    @height_limit.setter
    def height_limit(self, value):
        assert isinstance(value, (float, int)), \
            'Expected number for height_limit. Got {}.'.format(type(value))
        assert value > 0, 'The height_limit must be greater than zero.'
        self._height_limit = value

    @property
    def base_height(self):
        """Get a number for the absolute Z value of max/min acceptable height."""
        return self._geometry_mesh.center.z + self._height_limit \
            if self.solar_rights else \
            self._geometry_mesh.center.z - self._height_limit

    @property
    def solar_rights(self):
        """Get or set a boolean for whether a solar rights boundary is computed.

        If True, the evelope represents the boundary below which one can build]
        without shading the surrounding obstacles from any of the sun_vectors.
        If False, the envelope represents the boundary above which the one will
        have direct solar access to all of the input sun_vectors.
        """
        return self._solar_rights

    @solar_rights.setter
    def solar_rights(self, value):
        self._solar_rights = bool(value)

    @property
    def geometry_point2ds(self):
        """Get a list of Point2Ds for the vertices of the geometry_mesh."""
        return [Point2D(pt.x, pt.y) for pt in self._geometry_mesh]

    @property
    def obstacle_polygon2ds(self):
        """Get a list of Polygon2Ds for the obstacle boundaries in 2D space."""
        return [Polygon2D([Point2D(pt.x, pt.y) for pt in face.boundary])
                for face in self._obstacle_faces]

    @property
    def sun_vector2ds(self):
        """Get a list of Vector2Ds for the sun vectors in 2D space."""
        return [Vector2D(vec.x, vec.y) for vec in self._sun_vectors]

    @property
    def sun_vector2ds_reversed(self):
        """Get a list of Vector2Ds for the reversed sun vectors in 2D space."""
        return [Vector2D(-vec.x, -vec.y) for vec in self._sun_vectors]

    def envelope_mesh(self):
        """Compute a Mesh3D representing the solar envelope boundary."""
        # extract the relevant proprties from the input geometry
        pt2ds, poly2ds = self.geometry_point2ds, self.obstacle_polygon2ds
        vec2ds = self.sun_vector2ds if self.solar_rights else self.sun_vector2ds_reversed
        altitudes = self._sun_altitudes()
        obs_heights = [face[0].z for face in self._obstacle_faces]
        base_height = self.base_height

        # loop through the points to get the height of each one
        pt_heights = self._compute_point_heights(
            pt2ds, poly2ds, obs_heights, vec2ds, altitudes, base_height)

        # turn the mesh point heights into a full Mesh3D
        new_vertices = [Point3D(pt.x, pt.y, h) for pt, h in zip(pt2ds, pt_heights)]
        return Mesh3D(new_vertices, self._geometry_mesh.faces)

    def _sun_altitudes(self):
        """Get the altitudes of the sun_vectors in radians."""
        return [vec.angle(Vector3D(vec.x, vec.y, 0)) for vec in self._sun_vectors]

    def _compute_point_heights(
            self, pt2ds, obs_poly2ds, obs_heights, vec2ds, vec_altitudes, base_height):
        """Get Z heights for each of the points in the envelope mesh.

        Args:
            pt2ds: List of Point2D objects for each of the vertices of the mesh
            obs_poly2ds: List of Polygon2D objects for each of the obstacles.
            obs_heights: List of heights for each of the obstacles.
            vec2ds: List of Vector2Ds for each of the sun vectors. These should
                be reversed sun vectors when computing a solar collection.
            vec_altitudes: List of numbers for the altitude of each sun vector
                in radians.
            base_height: The starting height of the envelope points. This should
                be below the geometry_mesh for a collection boundary and above it
                for a rights boundary.
        """
        height_method = self._point_height_rights if self.solar_rights \
            else self._point_height_collection
        heights = []
        for point in pt2ds:
            pt_height = base_height
            for obs_poly, obs_h in zip(obs_poly2ds, obs_heights):
                for vec, alt in zip(vec2ds, vec_altitudes):
                    pt_height = height_method(
                        point, vec, alt, obs_poly, obs_h, pt_height)
            heights.append(pt_height)
        return heights

    @staticmethod
    def _point_height_collection(
            point, vector_2d, altitude, obstacle_poly, obstacle_height, default):
        """Get the height of a given 2D point and sun vector for a collection boundary.

        This will return None if the obstacle has no effect on the height.
        """
        int_pts = obstacle_poly.intersect_line_ray(Ray2D(point, vector_2d))
        if len(int_pts) != 0:
            pt_dist = min([point.distance_to_point(pt) for pt in int_pts])
            new_height = obstacle_height - pt_dist * math.tan(altitude)
            if new_height > default:
                return new_height
        return default

    @staticmethod
    def _point_height_rights(
            point, vector_2d, altitude, obstacle_poly, obstacle_height, default):
        """Get the height of a given 2D point and sun vector for a rights boundary.

        This will return None if the obstacle has no effect on the height.
        """
        int_pts = obstacle_poly.intersect_line_ray(Ray2D(point, vector_2d))
        if len(int_pts) != 0:
            pt_dist = min([point.distance_to_point(pt) for pt in int_pts])
            new_height = pt_dist * math.tan(altitude) + obstacle_height
            if new_height < default:
                return new_height
        return default

    def __repr__(self):
        """SolarEnvelope representation."""
        return "SolarEnvelope [{}]".format(self.geometry_mesh)



# coding=utf-8
"""Base data type."""
from __future__ import division

import os
import importlib
import re


class DataTypeBase(object):
    """Base class for data types.

    Args:
        name: Optional name for the type. Default is derived from the class name.

    Properties:
        * name
        * units
        * si_units
        * ip_units
        * min
        * max
        * abbreviation
        * unit_descr
        * point_in_time
        * cumulative
        * normalized_type
        * time_aggregated_type
        * time_aggregated_factor
    """
    _units = [None]
    _si_units = [None]
    _ip_units = [None]
    _min = float('-inf')
    _max = float('+inf')

    _abbreviation = ''
    _unit_descr = None
    _point_in_time = True
    _cumulative = False
    _normalized_type = None
    _time_aggregated_type = None
    _time_aggregated_factor = None

    _type_enumeration = None

    def __init__(self, name=None):
        """Initialize DataType.
        """
        self._name = name

    @classmethod
    def from_dict(cls, data):
        """Create a data type from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
                "name": ""  # data type name of the data type as a string
                "data_type": ""  # the class name of the data type as a string
            }
        """
        # load up all of the types if they aren't already loaded
        assert 'name' in data, 'Required keyword "name" is missing!'
        assert 'data_type' in data, 'Required keyword "data_type" is missing!'
        if cls._type_enumeration is None:
            cls._type_enumeration = _DataTypeEnumeration(import_modules=False)
        # load the data type from the dictionary
        if data['data_type'] == 'GenericType':
            assert 'base_unit' in data, \
                'Keyword "base_unit" is missing and is required for GenericType.'
            d_min = data['min'] if 'min' in data else float('-inf')
            d_max = data['max'] if 'max' in data else float('+inf')
            abbr = data['abbreviation'] if 'abbreviation' in data else None
            unit_descr = data['unit_descr'] if 'unit_descr' in data else None
            point_in_time = data['point_in_time'] if 'point_in_time' in data else True
            cumulative = data['cumulative'] if 'cumulative' in data else False
            return cls._type_enumeration._GENERICTYPE(
                data['name'], data['base_unit'], d_min, d_max, abbr,
                unit_descr, point_in_time, cumulative)
        elif data['data_type'] in cls._type_enumeration._TYPES:
            clss = cls._type_enumeration._TYPES[data['data_type']]
            if data['data_type'] == data['name'].title().replace(' ', ''):
                return clss()
            else:
                instance = clss()
                instance._name = data['name']
                return instance
        else:
            raise ValueError(
                'Data Type {} could not be recognized'.format(data['data_type']))

    @classmethod
    def from_string(cls, data_type_string):
        """Create a data type from a string.

        Args:
            data: A data type string.
        """
        # load up all of the types if they aren't already loaded
        if cls._type_enumeration is None:
            cls._type_enumeration = _DataTypeEnumeration(import_modules=False)
        # first, see if it is a standard data type
        d_type_class = data_type_string.title().replace(' ', '')
        if d_type_class in cls._type_enumeration._TYPES:
            clss = cls._type_enumeration._TYPES[d_type_class]
            return clss()
        # assume that it's a Generic data type
        return cls._type_enumeration._GENERICTYPE.from_string(data_type_string)

    def is_unit_acceptable(self, unit, raise_exception=True):
        """Check if a certain unit is acceptable for the data type.

        Args:
            unit: A text string representing the abbreviated unit.
            raise_exception: Set to True to raise an exception if not acceptable.
        """
        _is_acceptable = unit in self.units
        if _is_acceptable or not raise_exception:
            return _is_acceptable
        else:
            raise ValueError(
                '{0} is not an acceptable unit type for {1}. '
                'Choose from the following: {2}'.format(
                    unit, self.__class__.__name__, self.units
                )
            )

    def to_unit(self, values, unit, from_unit=None):
        """Return values converted to the unit given the input from_unit."""
        raise NotImplementedError(
            'to_unit is not implemented on %s' % self.__class__.__name__
        )

    def to_ip(self, values, from_unit=None):
        """Return values in IP and the units to which the values have been converted."""
        raise NotImplementedError(
            'to_ip is not implemented on %s' % self.__class__.__name__
        )

    def to_si(self, values, from_unit=None):
        """Return values in SI and the units to which the values have been converted."""
        raise NotImplementedError(
            'to_si is not implemented on %s' % self.__class__.__name__
        )

    def is_in_range(self, values, unit=None, raise_exception=True):
        """Check if a list of values is within physically/mathematically possible range.

        Args:
            values: A list of values.
            unit: The unit of the values.  If not specified, the default metric
                unit will be assumed.
            raise_exception: Set to True to raise an exception if not in range.
        """
        self._is_numeric(values)
        if unit is None or unit == self.units[0]:
            minimum = self.min
            maximum = self.max
        else:
            namespace = {'self': self}
            self.is_unit_acceptable(unit, True)
            min_statement = "self._{}_to_{}(self.min)".format(
                self._clean(self.units[0]), self._clean(unit))
            max_statement = "self._{}_to_{}(self.max)".format(
                self._clean(self.units[0]), self._clean(unit))
            minimum = eval(min_statement, namespace)
            maximum = eval(max_statement, namespace)

        for value in values:
            if value < minimum or value > maximum:
                if not raise_exception:
                    return False
                else:
                    raise ValueError(
                        '{0} should be between {1} and {2}. Got {3}'.format(
                            self.__class__.__name__, self.min, self.max, value
                        )
                    )
        return True

    def duplicate(self):
        """Return a copy of the data type."""
        return self.__class__(self.name)

    def to_dict(self):
        """Get data type as a dictionary."""
        return {
            'name': self.name,
            'data_type': self.__class__.__name__,
            'type': 'DataType'
        }

    def to_string(self):
        """Get data type as a string."""
        return self.name

    def _is_numeric(self, values):
        """Check to be sure values are numbers before doing numerical operations."""
        if len(values) > 0:
            assert isinstance(values[0], (float, int)), \
                "values must be numbers to perform math operations. Got {}".format(
                    type(values[0]))
        return True

    def _to_unit_base(self, base_unit, values, unit, from_unit):
        """Return values in a given unit given the input from_unit."""
        self._is_numeric(values)
        namespace = {'self': self, 'values': values}
        if not from_unit == base_unit:
            self.is_unit_acceptable(from_unit, True)
            statement = '[self._{}_to_{}(val) for val in values]'.format(
                self._clean(from_unit), self._clean(base_unit))
            values = eval(statement, namespace)
            namespace['values'] = values
        if not unit == base_unit:
            self.is_unit_acceptable(unit, True)
            statement = '[self._{}_to_{}(val) for val in values]'.format(
                self._clean(base_unit), self._clean(unit))
            values = eval(statement, namespace)
        return values

    def _clean(self, unit):
        """Clean out special characters from unit abbreviations."""
        return unit.replace(
            '/', '_').replace(
                '-', '').replace(
                    ' ', '').replace(
                        '%', 'pct')

    @property
    def name(self):
        """The full name of the data type as a string."""
        if self._name is None:
            return re.sub(r"(?<=\w)([A-Z])", r" \1", self.__class__.__name__)
        else:
            return self._name

    @property
    def units(self):
        """A tuple of all acceptable units of the data type as abbreviated text.

        The first item of the list should be the standard SI unit.
        The second item of the list should be the standard IP unit (if it exists).
        The rest of the list can be any other acceptable units.
        (eg. [C, F, K])
        """
        return self._units

    @property
    def si_units(self):
        """A tuple of acceptable si_units for the data type."""
        return self._si_units

    @property
    def ip_units(self):
        """A tuple of acceptable ip_units for the data type."""
        return self._ip_units

    @property
    def min(self):
        """Number for the lower limit for the data type.

        Values below this limit should be physically or mathematically impossible.
        """
        return self._min

    @property
    def max(self):
        """Number for the upper limit for the data type.

        Values above this limit should be physically or mathematically impossible.
        """
        return self._max

    @property
    def abbreviation(self):
        """An optional abbreviation for the data type as text.

        (eg. 'UTCI' for Universal Thermal Climate Index).
        This can also be a letter that represents the data type in a formula.
        (eg. 'A' for Area; 'P' for Pressure)
        """
        return self._abbreviation

    @property
    def unit_descr(self):
        """A dictionary that matches numerical values to text categories.

        This will be None if there are no text categories that the data type can be
        mapped to. (eg. -1 = Cold, 0 = Neutral, +1 = Hot) (eg. 0 = False, 1 = True).
        """
        return self._unit_descr

    @property
    def point_in_time(self):
        """Boolean to note whether data type is for a single instant in time.

        If False, the data type is meant to represent an average or accumulation
        over time whenever found in an array of time series data.
        (True Examples: Temperature, WindSpeed)
        (False Examples: Energy, Radiation, Illuminance)
        """
        return self._point_in_time

    @property
    def cumulative(self):
        """Boolean to note if data type can be summed over time to yield meaningful data.

        If False, this data type can only be averaged over time to be meaningful.
        Note that cumulative cannot be True when point_in_time is also True.
        (False Examples: Temperature, Irradiance, Illuminance)
        (True Examples: Energy, Radiation)
        """
        return self._cumulative

    @property
    def normalized_type(self):
        """A data type object representing the area-normalized version of this data type.

        This will be None if the data type cannot be normalized per unit area to
        yield a meaningful data type.
        """
        return self._normalized_type

    @property
    def time_aggregated_type(self):
        """A data type object representing the time-aggregated version of this data type.

        This will be None if the data type cannot be aggregated per unit time to
        yield a meaningful data type.
        """
        return self._time_aggregated_type

    @property
    def time_aggregated_factor(self):
        """A number to convert to the base unit of the type to the time aggregated unit.

        The factor assumes that the data is aggregated over one hour. This will be
        None if the data type cannot be aggregated per unit time to yield a
        meaningful data type.
        """
        return self._time_aggregated_factor

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __key(self):
        return (
            self._name, self._units, self._si_units, self._ip_units, self._min,
            self._max, self._abbreviation, self._unit_descr, self._point_in_time,
            self._cumulative, self._normalized_type, self._time_aggregated_type
        )

    def __eq__(self, other):
        return isinstance(other, DataTypeBase) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        """Return Ladybug data type as a string."""
        return self.name


class _DataTypeEnumeration(object):
    """Enumerates all data types, base types, and units."""
    _TYPES = {}
    _BASETYPES = {}
    _UNITS = {}
    _GENERICTYPE = None

    def __init__(self, import_modules=True):
        if import_modules:
            self._import_modules()

        for clss in DataTypeBase.__subclasses__():
            if clss.__name__ != 'GenericType':
                self._TYPES[clss.__name__] = clss
                self._BASETYPES[clss.__name__] = clss
                self._UNITS[clss.__name__] = clss._units
                for subclss in self._all_subclasses(clss):
                    self._TYPES[subclss.__name__] = subclss
            else:
                self._GENERICTYPE = clss

    @property
    def types(self):
        """A tuple indicating all currently supported data types."""
        return tuple(sorted(self._TYPES.keys()))

    @property
    def base_types(self):
        """A tuple indicating all base types.

        Base types are the data types on which unit systems are defined.
        """
        return tuple(sorted(self._BASETYPES.keys()))

    @property
    def units(self):
        """A dictionary containing all currently supported units.

        The keys of this dictionary are the base types (eg. 'Temperature').
        """
        return self._UNITS

    @property
    def types_dict(self):
        """A dictionary containing pointers to the classes of each data type.

        The keys of this dictionary are the data types.
        """
        return self._TYPES

    def _import_modules(self):
        root_dir = os.path.dirname(__file__)
        modules = os.listdir(os.path.dirname(__file__))
        modules = [os.path.join(root_dir, mod) for mod in modules]
        importable = ['.{}'.format(os.path.basename(f)[:-3]) for f in modules
                      if os.path.isfile(f) and f.endswith('.py')
                      and not f.endswith('__init__.py')
                      and not f.endswith('base.py')]
        for mod in importable:
            importlib.import_module(mod, 'ladybug.datatype')

    def _all_subclasses(self, clss):
        return set(clss.__subclasses__()).union(
            [s for c in clss.__subclasses__() for s in self._all_subclasses(c)])



# coding=utf-8
from __future__ import division

import math
import os
from copy import deepcopy

from ladybug_geometry.geometry3d.pointvector import Vector3D

from .analysisperiod import AnalysisPeriod
from .datacollection import HourlyContinuousCollection, HourlyDiscontinuousCollection
from .datatype.energyflux import Irradiance, GlobalHorizontalIrradiance, \
    DirectNormalIrradiance, DiffuseHorizontalIrradiance, DirectHorizontalIrradiance
from .datatype.illuminance import GlobalHorizontalIlluminance, \
    DirectNormalIlluminance, DiffuseHorizontalIlluminance
from .datatype.luminance import ZenithLuminance
from .dt import DateTime, Time
from .epw import EPW
from .futil import write_to_file
from .header import Header
from .location import Location
from .skymodel import ashrae_revised_clear_sky, ashrae_clear_sky, \
    zhang_huang_solar_split, estimate_illuminance_from_irradiance
from .stat import STAT
from .sunpath import Sunpath

try:  # python 2
    from itertools import izip as zip
    readmode = 'rb'
    writemode = 'wb'
except ImportError:  # python 3
    xrange = range
    readmode = 'r'
    writemode = 'w'
    xrange = range


class Wea(object):
    """A WEA object containing hourly or sub-hourly solar irradiance.

    This object and its corresponding .wea file type is what the Radiance gendaymtx
    function uses to generate the sky.

    Args:
        location: Ladybug location object.
        direct_normal_irradiance: A HourlyContinuousCollection or a
            HourlyDiscontinuousCollection for direct normal irradiance. The
            collection must be aligned with the diffuse_horizontal_irradiance.
        diffuse_horizontal_irradiance: A HourlyContinuousCollection or a
            HourlyDiscontinuousCollection for diffuse horizontal irradiance, The
            collection must be aligned with the direct_normal_irradiance.

    Properties:
        * location
        * direct_normal_irradiance
        * diffuse_horizontal_irradiance
        * direct_horizontal_irradiance
        * global_horizontal_irradiance
        * enforce_on_hour
        * datetimes
        * hoys
        * analysis_period
        * timestep
        * is_leap_year
        * is_continuous
        * is_annual
        * header
    """
    __slots__ = \
        ('_timestep', '_is_leap_year', '_location', 'metadata', '_enforce_on_hour',
         '_direct_normal_irradiance', '_diffuse_horizontal_irradiance')

    def __init__(
        self, location, direct_normal_irradiance, diffuse_horizontal_irradiance
    ):
        """Create a Wea object."""
        # Check that input collections are of the right type and aligned to each other
        acceptable_colls = (HourlyContinuousCollection, HourlyDiscontinuousCollection)
        for coll in (direct_normal_irradiance, diffuse_horizontal_irradiance):
            assert isinstance(coll, acceptable_colls), 'Input irradiance data for ' \
                'Wea must be an hourly data collection. Got {}.'.format(type(coll))
        assert direct_normal_irradiance.is_collection_aligned(
            diffuse_horizontal_irradiance), 'Wea direct normal and diffuse horizontal ' \
            'irradiance collections must be aligned with one another.'

        # assign the location, irradiance, metadata, timestep and leap year
        self._enforce_on_hour = False  # False by default
        self.location = location
        self._direct_normal_irradiance = direct_normal_irradiance
        self._diffuse_horizontal_irradiance = diffuse_horizontal_irradiance
        self.metadata = {'source': location.source, 'country': location.country,
                         'city': location.city}
        self._timestep = direct_normal_irradiance.header.analysis_period.timestep
        self._is_leap_year = direct_normal_irradiance.header.analysis_period.is_leap_year

    @classmethod
    def from_annual_values(
        cls, location, direct_normal_irradiance, diffuse_horizontal_irradiance,
        timestep=1, is_leap_year=False
    ):
        """Create an annual Wea from an array of irradiance values.

        Args:
            location: Ladybug location object.
            direct_normal_irradiance: An array of values for direct normal irradiance.
                The length of this list should be same as diffuse_horizontal_irradiance
                and should represent an entire year of values at the input timestep.
            diffuse_horizontal_irradiance: A HourlyContinuousCollection or a
                HourlyDiscontinuousCollection for diffuse horizontal irradiance, The
                collection must be aligned with the direct_normal_irradiance.
            timestep: An integer to set the number of time steps per hour.
                Default is 1 for one value per hour.
            is_leap_year: A boolean to indicate if values are for a leap
                year. (Default: False).
        """
        metadata = {'source': location.source, 'country': location.country,
                    'city': location.city}
        dnr, dhr = cls._get_data_collections(
            direct_normal_irradiance, diffuse_horizontal_irradiance,
            metadata, timestep, is_leap_year)
        return cls(location, dnr, dhr)

    @classmethod
    def from_dict(cls, data):
        """ Create Wea from a dictionary

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "Wea",
            "location": {},  # ladybug location dictionary
            "direct_normal_irradiance": [],  # direct normal irradiance values
            "diffuse_horizontal_irradiance": [],  # diffuse horizontal irradiance values
            "timestep": 1,  # optional timestep between measurements
            "is_leap_year": False,  # optional boolean for leap year
            "datetimes": []  # array of datetime arrays; only required when not annual
            }
        """
        # check for the required keys
        required_keys = ('type', 'location', 'direct_normal_irradiance',
                         'diffuse_horizontal_irradiance')
        for key in required_keys:
            assert key in data, 'Required key "{}" is missing!'.format(key)
        assert data['type'] == 'Wea', \
            'Expected Wea dictionary. Got {}.'.format(data['type'])

        # set the optional properties
        timestep = data['timestep'] if 'timestep' in data else 1
        is_leap_year = data['is_leap_year'] if 'is_leap_year' in data else False

        # correctly interpret the datetimes to create correct analysis periods
        continuous = True
        if 'datetimes' in data and data['datetimes'] is not None:
            st_dt = DateTime.from_array(data['datetimes'][0])
            end_dt = DateTime.from_array(data['datetimes'][-1])
            if st_dt.leap_year is not is_leap_year:
                st_dt = DateTime(st_dt.month, st_dt.day, st_dt.hour, is_leap_year)
                end_dt = DateTime(end_dt.month, end_dt.day, end_dt.hour, is_leap_year)
            a_per = AnalysisPeriod.from_start_end_datetime(st_dt, end_dt, timestep)
            if a_per.st_hour != 0 or a_per.end_hour != 23:
                continuous = False
            if len(a_per) != len(data['direct_normal_irradiance']):
                a_per = AnalysisPeriod(timestep=timestep, is_leap_year=is_leap_year)
                continuous = False
        else:  # assume it is annual continuous data
            a_per = AnalysisPeriod(timestep=timestep, is_leap_year=is_leap_year)

        # serialize the location and data collections
        location = Location.from_dict(data['location'])
        dni_head = Header(DirectNormalIrradiance(), 'W/m2', a_per)
        dhi_head = Header(DiffuseHorizontalIrradiance(), 'W/m2', a_per)
        if continuous:
            dni = HourlyContinuousCollection(
                dni_head, data['direct_normal_irradiance'])
            dhi = HourlyContinuousCollection(
                dhi_head, data['diffuse_horizontal_irradiance'])
        else:
            datetimes = [DateTime.from_array(dat) for dat in data['datetimes']]
            dni = HourlyDiscontinuousCollection(
                dni_head, data['direct_normal_irradiance'], datetimes)
            dhi = HourlyDiscontinuousCollection(
                dhi_head, data['diffuse_horizontal_irradiance'], datetimes)

        return cls(location, dni, dhi)

    @classmethod
    def from_file(cls, wea_file, timestep=1, is_leap_year=False):
        """Create Wea object from a .wea file.

        Args:
            wea_file:Full path to .wea file.
            timestep: An optional integer to set the number of time steps per hour.
                Default is 1 for one value per hour. If the wea file has a time step
                smaller than an hour, adjust this input accordingly.
            is_leap_year: A boolean to indicate if values are for a leap
                year. (Default: False).
        """
        assert os.path.isfile(wea_file), 'Failed to find {}'.format(wea_file)
        with open(wea_file, readmode) as weaf:
            location = cls._parse_wea_header(weaf, wea_file)
            # parse irradiance values
            dir_norm_irr = []
            dif_horiz_irr = []
            dt_arr = []
            for line in weaf:
                vals = line.split()
                dir_norm_irr.append(float(vals[-2]))
                dif_horiz_irr.append(float(vals[-1]))
                dt_arr.append([int(vals[0]), int(vals[1]), float(vals[2])])

        # interpret datetimes to create data collections with correct analysis periods
        continuous = True
        st_dt = DateTime.from_array([dt_arr[0][0], dt_arr[0][1], int(dt_arr[0][2])])
        end_dt = DateTime.from_array([dt_arr[-1][0], dt_arr[-1][1], int(dt_arr[-1][2])])
        if st_dt.leap_year is not is_leap_year:
            st_dt = DateTime(st_dt.month, st_dt.day, st_dt.hour, is_leap_year)
            end_dt = DateTime(end_dt.month, end_dt.day, end_dt.hour, is_leap_year)
        a_per = AnalysisPeriod.from_start_end_datetime(st_dt, end_dt, timestep)
        if a_per.st_hour != 0 or a_per.end_hour != 23:  # potential continuous time slice
            continuous = False
        if len(a_per) != len(dir_norm_irr):  # true discontinuous data
            a_per = AnalysisPeriod(timestep=timestep, is_leap_year=is_leap_year)
            continuous = False

        # serialize the data collections
        metadata = {'city': location.city}
        dni_head = Header(DirectNormalIrradiance(), 'W/m2', a_per, metadata)
        dhi_head = Header(DiffuseHorizontalIrradiance(), 'W/m2', a_per, metadata)
        if continuous:
            dni = HourlyContinuousCollection(dni_head, dir_norm_irr)
            dhi = HourlyContinuousCollection(dhi_head, dif_horiz_irr)
        else:
            if timestep == 1:
                datetimes = [DateTime(d[0], d[1], int(d[2])) for d in dt_arr]
            else:
                datetimes = []
                for d in dt_arr:
                    tim = Time.from_mod(int(d[2] * 60))
                    datetimes.append(DateTime(d[0], d[1], tim.hour, tim.minute))
            dni = HourlyDiscontinuousCollection(dni_head, dir_norm_irr, datetimes)
            dhi = HourlyDiscontinuousCollection(dhi_head, dif_horiz_irr, datetimes)
            dni = dni.validate_analysis_period()
            dhi = dhi.validate_analysis_period()

        return cls(location, dni, dhi)

    @classmethod
    def from_daysim_file(cls, wea_file, timestep=1, is_leap_year=False):
        """Create Wea object from a .wea file produced by DAYSIM.

        Note that this method is only required when the .wea file generated from
        DAYSIM has a timestep greater than 1, which results in the file using
        times of day greater than 23:59. DAYSIM weas with a timestep of 1 can
        use the from_file method without issues.

        Args:
            wea_file:Full path to .wea file.
            timestep: An optional integer to set the number of time steps per hour.
                Default is 1 for one value per hour.
            is_leap_year: A boolean to indicate if values are for a leap
                year. (Default: False).
        """
        # parse in the data
        assert os.path.isfile(wea_file), 'Failed to find {}'.format(wea_file)
        with open(wea_file, readmode) as weaf:
            location = cls._parse_wea_header(weaf, wea_file)
            # parse irradiance values
            dir_norm_irr = []
            dif_horiz_irr = []
            for line in weaf:
                dirn, difh = [int(v) for v in line.split()[-2:]]
                dir_norm_irr.append(dirn)
                dif_horiz_irr.append(difh)

        # move the last half hour of data to the start of the file
        if timestep != 1:
            shift = -int(timestep / 2)
            dir_norm_irr = dir_norm_irr[shift:] + dir_norm_irr[:shift]
            dif_horiz_irr = dif_horiz_irr[shift:] + dif_horiz_irr[:shift]

        return cls.from_annual_values(
            location, dir_norm_irr, dif_horiz_irr, timestep, is_leap_year)

    @classmethod
    def from_epw_file(cls, epw_file, timestep=1):
        """Create a wea object using the solar irradiance values in an epw file.

        Args:
            epw_file: Full path to epw weather file.
            timestep: An optional integer to set the number of time steps per hour.
                Default is 1 for one value per hour. Note that this input
                will only do a linear interpolation over the data in the EPW
                file.  While such linear interpolations are suitable for most
                thermal simulations, where thermal lag "smooths over" the effect
                of momentary increases in solar energy, it is not recommended
                for daylight simulations, where momentary increases in solar
                energy can mean the difference between glare and visual comfort.
        """
        epw = EPW(epw_file)
        direct_normal, diffuse_horizontal = \
            cls._get_data_collections(epw.direct_normal_radiation.values,
                                      epw.diffuse_horizontal_radiation.values,
                                      epw.metadata, 1, epw.is_leap_year)
        if timestep != 1:
            print("Note: timesteps greater than 1 on epw-generated Weas \n"
                  "are suitable for thermal models but are not recommended \n"
                  "for daylight models.")
            # interpolate the data
            direct_normal = direct_normal.interpolate_to_timestep(timestep)
            diffuse_horizontal = diffuse_horizontal.interpolate_to_timestep(timestep)
            # create sunpath to check if the sun is up at a given timestep
            sp = Sunpath.from_location(epw.location)
            # add correct values to the empty data collection
            for i, dt in enumerate(cls._get_datetimes(timestep, epw.is_leap_year)):
                # set irradiance values to 0 when the sun is not up
                sun = sp.calculate_sun_from_date_time(dt)
                if sun.altitude < 0:
                    direct_normal[i] = 0
                    diffuse_horizontal[i] = 0

        return cls(epw.location, direct_normal, diffuse_horizontal)

    @classmethod
    def from_stat_file(cls, statfile, timestep=1, is_leap_year=False, use_2017=False):
        """Create an ASHRAE Revised Clear Sky Wea object from data in .stat file.

        The .stat file must have monthly sky optical depths within it in order to
        create a Wea this way.

        Args:
            statfile: Full path to the .stat file.
            timestep: An optional integer to set the number of time steps per
                hour. Default is 1 for one value per hour.
            is_leap_year: A boolean to indicate if values are for a leap
                year. (Default: False).
            use_2017: A boolean to indicate whether the version of the ASHRAE Tau
                model that should be the revised version published in 2017 (True)
                or the original one published in 2009 (False). (Default: False).
        """
        stat = STAT(statfile)

        # check to be sure the stat file does not have missing tau values
        def check_missing(opt_data, data_name):
            if opt_data == []:
                raise ValueError('Stat file contains no optical data.')
            for i, x in enumerate(opt_data):
                if x is None:
                    raise ValueError(
                        'Missing optical depth data for {} at month {}'.format(
                            data_name, i)
                    )
        check_missing(stat.monthly_tau_beam, 'monthly_tau_beam')
        check_missing(stat.monthly_tau_diffuse, 'monthly_tau_diffuse')

        return cls.from_ashrae_revised_clear_sky(
            stat.location, stat.monthly_tau_beam, stat.monthly_tau_diffuse,
            timestep, is_leap_year, use_2017)

    @classmethod
    def from_ashrae_revised_clear_sky(cls, location, monthly_tau_beam,
                                      monthly_tau_diffuse, timestep=1,
                                      is_leap_year=False, use_2017=False):
        """Create a wea object representing an ASHRAE Revised Clear Sky ("Tau Model")

        ASHRAE Revised Clear Skies are intended to determine peak solar load
        and sizing parameters for HVAC systems.  The revised clear sky is
        currently the default recommended sky model used to autosize HVAC
        systems in EnergyPlus. For more information on the ASHRAE Revised Clear
        Sky model, see the EnergyPlus Engineering Reference:
        https://bigladdersoftware.com/epx/docs/23-2/engineering-reference/climate-calculations.html

        Args:
            location: Ladybug location object.
            monthly_tau_beam: A list of 12 float values indicating the beam
                optical depth of the sky at each month of the year.
            monthly_tau_diffuse: A list of 12 float values indicating the
                diffuse optical depth of the sky at each month of the year.
            timestep: An optional integer to set the number of time steps per
                hour. Default is 1 for one value per hour.
            is_leap_year: A boolean to indicate if values are for a leap
                year. (Default: False).
            use_2017: A boolean to indicate whether the version of the ASHRAE Tau
                model that should be the revised version published in 2017 (True)
                or the original one published in 2009 (False). (Default: False).
        """
        # extract metadata
        metadata = {'source': location.source, 'country': location.country,
                    'city': location.city}

        # create sunpath and get altitude at every timestep of the year
        sp = Sunpath.from_location(location)
        sp.is_leap_year = is_leap_year
        altitudes = [[] for i in range(12)]
        dates = cls._get_datetimes(timestep, is_leap_year)
        for t_date in dates:
            sun = sp.calculate_sun_from_date_time(t_date)
            altitudes[sun.datetime.month - 1].append(sun.altitude)

        # run all of the months through the ashrae_revised_clear_sky model
        direct_norm, diffuse_horiz = [], []
        for i_mon, alt_list in enumerate(altitudes):
            dir_norm_rad, dif_horiz_rad = ashrae_revised_clear_sky(
                alt_list, monthly_tau_beam[i_mon], monthly_tau_diffuse[i_mon], use_2017)
            direct_norm.extend(dir_norm_rad)
            diffuse_horiz.extend(dif_horiz_rad)

        direct_norm_rad, diffuse_horiz_rad = \
            cls._get_data_collections(direct_norm, diffuse_horiz,
                                      metadata, timestep, is_leap_year)

        return cls(location, direct_norm_rad, diffuse_horiz_rad)

    @classmethod
    def from_ashrae_clear_sky(cls, location, sky_clearness=1, timestep=1,
                              is_leap_year=False):
        """Create a wea object representing an original ASHRAE Clear Sky.

        The original ASHRAE Clear Sky is intended to determine peak solar load
        and sizing parameters for HVAC systems.  It is not the sky model
        currently recommended by ASHRAE since it usually overestimates the
        amount of solar irradiance in comparison to the newer ASHRAE Revised
        Clear Sky ("Tau Model"). However, the original model here is still
        useful for cases where monthly optical depth values are not known. For
        more information on the ASHRAE Clear Sky model, see the EnergyPlus
        Engineering Reference:
        https://bigladdersoftware.com/epx/docs/8-9/engineering-reference/climate-calculations.html

        Args:
            location: Ladybug location object.
            sky_clearness: A factor that will be multiplied by the output of
                the model. This is to help account for locations where clear,
                dry skies predominate (e.g., at high elevations) or,
                conversely, where hazy and humid conditions are frequent. See
                Threlkeld and Jordan (1958) for recommended values. Typical
                values range from 0.95 to 1.05 and are usually never more
                than 1.2. Default is set to 1.0.
            timestep: An optional integer to set the number of time steps per
                hour. Default is 1 for one value per hour.
            is_leap_year: A boolean to indicate if values are for a leap
                year. (Default: False).
        """
        # extract metadata
        metadata = {'source': location.source, 'country': location.country,
                    'city': location.city}

        # create sunpath and get altitude at every timestep of the year
        sp = Sunpath.from_location(location)
        sp.is_leap_year = is_leap_year
        altitudes = [[] for i in range(12)]
        dates = cls._get_datetimes(timestep, is_leap_year)
        for t_date in dates:
            sun = sp.calculate_sun_from_date_time(t_date)
            altitudes[sun.datetime.month - 1].append(sun.altitude)

        # compute hourly direct normal and diffuse horizontal irradiance
        direct_norm, diffuse_horiz = [], []
        for i_mon, alt_list in enumerate(altitudes):
            dir_norm_rad, dif_horiz_rad = ashrae_clear_sky(
                alt_list, i_mon + 1, sky_clearness)
            direct_norm.extend(dir_norm_rad)
            diffuse_horiz.extend(dif_horiz_rad)

        direct_norm_rad, diffuse_horiz_rad = \
            cls._get_data_collections(direct_norm, diffuse_horiz,
                                      metadata, timestep, is_leap_year)

        return cls(location, direct_norm_rad, diffuse_horiz_rad)

    @classmethod
    def from_zhang_huang_solar(cls, location, cloud_cover, relative_humidity,
                               dry_bulb_temperature, wind_speed,
                               atmospheric_pressure=None, use_disc=False):
        """Create a Wea object from climate data using the Zhang-Huang model.

        The Zhang-Huang solar model was developed to estimate solar
        irradiance for weather stations that lack such values, which are
        typically colleted with a pyranometer. Using total cloud cover,
        dry-bulb temperature, relative humidity, and wind speed as
        inputs the Zhang-Huang estimates global horizontal irradiance
        by means of a regression model across these variables.
        For more information on the Zhang-Huang model, see the
        EnergyPlus Engineering Reference:
        https://bigladdersoftware.com/epx/docs/8-7/engineering-reference/climate-calculations.html#zhang-huang-solar-model

        Args:
            location: Ladybug location object.
            cloud_cover: A hourly continuous data collection with values for the
                fraction of the sky dome covered in clouds (0 = clear;
                1 = completely overcast).
            relative_humidity: A hourly continuous data collection with values for
                the relative humidity in percent.
            dry_bulb_temperature: A hourly continuous data collection with values
                for the dry bulb temperature in degrees Celsius.
            wind_speed: A hourly continuous data collection with values for the
                wind speed in meters per second.
            atmospheric_pressure: An optional hourly continuous data collection
                with values for the atmospheric pressure in Pa. If None, pressure
                at sea level will be used (101325 Pa). (Default: None)
            use_disc: Boolean to note whether the original DISC model as opposed to the
                newer and more accurate DIRINT model. (Default: False).
        """
        # Check that input collections are of the right type and aligned to each other
        colls = (cloud_cover, relative_humidity, dry_bulb_temperature, wind_speed)
        for coll in colls:
            assert isinstance(coll, HourlyContinuousCollection), 'Input data for Zhang' \
                '-Huang Wea must be an hourly continuous. Got {}.'.format(type(coll))
        assert cloud_cover.are_collections_aligned(colls), 'Zhang-Huang Wea input ' \
            'data collections must be aligned with one another.'

        # check atmospheric_pressure input and generate default if None
        if atmospheric_pressure is not None:
            assert cloud_cover.is_collection_aligned(atmospheric_pressure), \
                'length pf atmospheric_pressure must match the other input collections.'
            atm_pressure = atmospheric_pressure.values
        else:
            atm_pressure = [101325] * len(cloud_cover)

        # initiate sunpath based on location
        sp = Sunpath.from_location(location)
        sp.is_leap_year = cloud_cover.header.analysis_period.is_leap_year
        a_per = cloud_cover.header.analysis_period

        # calculate parameters needed for zhang-huang irradiance
        date_times = []
        altitudes = []
        doys = []
        dry_bulb_t3_hrs = []
        for count, t_date in enumerate(cloud_cover.datetimes):
            date_times.append(t_date)
            sun = sp.calculate_sun_from_date_time(t_date)
            altitudes.append(sun.altitude)
            doys.append(sun.datetime.doy)
            dry_bulb_t3_hrs.append(dry_bulb_temperature[count - (3 * a_per.timestep)])

        # calculate zhang-huang irradiance
        dir_ir, diff_ir = zhang_huang_solar_split(
            altitudes, doys, cloud_cover.values, relative_humidity.values,
            dry_bulb_temperature.values, dry_bulb_t3_hrs, wind_speed.values,
            atm_pressure, use_disc)

        # assemble the results into DataCollections
        metadata = {'source': location.source, 'country': location.country,
                    'city': location.city}
        dni_head = Header(DirectNormalIrradiance(), 'W/m2', a_per, metadata)
        dhi_head = Header(DiffuseHorizontalIrradiance(), 'W/m2', a_per, metadata)
        dni = HourlyContinuousCollection(dni_head, dir_ir)
        dhi = HourlyContinuousCollection(dhi_head, diff_ir)
        return cls(location, dni, dhi)

    @property
    def enforce_on_hour(self):
        """Get or set a boolean for whether datetimes occur on the hour.

        By default, datetimes will be on the half-hour whenever the Wea has a
        timestep of 1, which aligns best with epw data. Setting this property
        to True will force the datetimes to be on the hour. Note that this
        property has no effect when the Wea timestep is not 1.
        """
        return self._enforce_on_hour

    @enforce_on_hour.setter
    def enforce_on_hour(self, value):
        self._enforce_on_hour = bool(value)

    @property
    def datetimes(self):
        """Get the datetimes in the Wea as a tuple of datetimes."""
        if self.timestep == 1 and not self._enforce_on_hour:
            return tuple(dt.add_minute(30) for dt in
                         self.direct_normal_irradiance.datetimes)
        else:
            return self.direct_normal_irradiance.datetimes

    @property
    def hoys(self):
        """Get the hours of the year in Wea as a tuple of floats."""
        return tuple(dt.hoy for dt in self.datetimes)

    @property
    def analysis_period(self):
        """Get an AnalysisPeriod for the Wea data."""
        return self._direct_normal_irradiance.header.analysis_period

    @property
    def timestep(self):
        """Get the timesteps per hour of the Wea as an integer."""
        return self._timestep

    @property
    def is_leap_year(self):
        """Get a boolean for whether the irradiance data is for a leap year."""
        return self._is_leap_year

    @property
    def is_continuous(self):
        """Get a boolean for whether the irradiance data is continuous."""
        return isinstance(self._direct_normal_irradiance, HourlyContinuousCollection)

    @property
    def is_annual(self):
        """Get a boolean for whether the irradiance data is for an entire year."""
        return self.is_continuous and self.analysis_period.is_annual

    @property
    def header(self):
        """Get the Wea header as a string."""
        return "place %s\n" % self.location.city + \
            "latitude %.2f\n" % self.location.latitude + \
            "longitude %.2f\n" % -self.location.longitude + \
            "time_zone %d\n" % (-self.location.time_zone * 15) + \
            "site_elevation %.1f\n" % self.location.elevation + \
            "weather_data_file_units 1\n"

    @property
    def location(self):
        """Get or set a Ladybug Location object for the Wea."""
        return self._location

    @location.setter
    def location(self, value):
        assert isinstance(value, Location), \
            'Wea.location data must be a Ladybug Location. Got {}'.format(type(value))
        self._location = value

    @property
    def direct_normal_irradiance(self):
        """Get or set a hourly data collection for the direct normal irradiance."""
        return self._direct_normal_irradiance

    @direct_normal_irradiance.setter
    def direct_normal_irradiance(self, data):
        acceptable_colls = (HourlyContinuousCollection, HourlyDiscontinuousCollection)
        assert isinstance(data, acceptable_colls), 'Input irradiance data for ' \
            'Wea must be an hourly data collection. Got {}.'.format(type(data))
        assert data.is_collection_aligned(self.diffuse_horizontal_irradiance), \
            'Wea direct normal and diffuse horizontal ' \
            'irradiance collections must be aligned with one another.'
        assert isinstance(data.header.data_type, DirectNormalIrradiance), \
            'direct_normal_irradiance data type must be' \
            'DirectNormalIrradiance. Got {}'.format(type(data.header.data_type))
        self._direct_normal_irradiance = data

    @property
    def diffuse_horizontal_irradiance(self):
        """Get or set a hourly data collection for the diffuse horizontal irradiance."""
        return self._diffuse_horizontal_irradiance

    @diffuse_horizontal_irradiance.setter
    def diffuse_horizontal_irradiance(self, data):
        acceptable_colls = (HourlyContinuousCollection, HourlyDiscontinuousCollection)
        assert isinstance(data, acceptable_colls), 'Input irradiance data for ' \
            'Wea must be an hourly data collection. Got {}.'.format(type(data))
        assert data.is_collection_aligned(self.direct_normal_irradiance), \
            'Wea direct normal and diffuse horizontal ' \
            'irradiance collections must be aligned with one another.'
        assert isinstance(data.header.data_type, DiffuseHorizontalIrradiance), \
            'direct_normal_irradiance data type must be' \
            'DiffuseHorizontalIrradiance. Got {}'.format(type(data.header.data_type))
        self._diffuse_horizontal_irradiance = data

    @property
    def global_horizontal_irradiance(self):
        """Get a data collection for the global horizontal irradiance."""
        header_ghr = Header(data_type=GlobalHorizontalIrradiance(),
                            unit='W/m2',
                            analysis_period=self.analysis_period,
                            metadata=self.metadata)
        glob_horiz = []
        sp = Sunpath.from_location(self.location)
        sp.is_leap_year = self.is_leap_year
        for dt, dnr, dhr in zip(self.datetimes, self.direct_normal_irradiance,
                                self.diffuse_horizontal_irradiance):
            sun = sp.calculate_sun_from_date_time(dt)
            glob_horiz.append(dhr + dnr * math.sin(math.radians(sun.altitude)))
        return self._aligned_collection(header_ghr, glob_horiz)

    @property
    def direct_horizontal_irradiance(self):
        """Get a data collection for the direct irradiance on a horizontal surface.

        Note that this is different from the direct_normal_irradiance needed
        to construct a Wea, which is NORMAL and not HORIZONTAL.
        """
        header_dhr = Header(data_type=DirectHorizontalIrradiance(),
                            unit='W/m2',
                            analysis_period=self.analysis_period,
                            metadata=self.metadata)
        direct_horiz = []
        sp = Sunpath.from_location(self.location)
        sp.is_leap_year = self.is_leap_year
        for dt, dnr in zip(self.datetimes, self.direct_normal_irradiance):
            sun = sp.calculate_sun_from_date_time(dt)
            direct_horiz.append(dnr * math.sin(math.radians(sun.altitude)))
        return self._aligned_collection(header_dhr, direct_horiz)

    def filter_by_pattern(self, pattern):
        """Create a new filtered Wea from this Wea using a list of booleans.

        Args:
            pattern: An array of True/False values. This array should usually
                have a length matching the number of irradiance values in the Wea
                but it can also be a pattern to be repeated over the data.

        Returns:
            A new Wea filtered by the analysis period.
        """
        return Wea(
            self.location,
            self.direct_normal_irradiance.filter_by_pattern(pattern),
            self.diffuse_horizontal_irradiance.filter_by_pattern(pattern))

    def filter_by_analysis_period(self, analysis_period):
        """Create a new filtered Wea from this Wea based on an analysis period.

        Args:
           analysis period: A Ladybug analysis period.

        Returns:
            A new Wea filtered by the analysis period.
        """
        return Wea(
            self.location,
            self.direct_normal_irradiance.filter_by_analysis_period(analysis_period),
            self.diffuse_horizontal_irradiance.filter_by_analysis_period(analysis_period)
        )

    def filter_by_hoys(self, hoys):
        """Create a new filtered Wea from this Wea using a list of hours of the year.

        Args:
           hoys: A List of hours of the year 0..8759.

        Returns:
            A new Wea with filtered data.
        """
        return Wea(
            self.location,
            self.direct_normal_irradiance.filter_by_hoys(hoys),
            self.diffuse_horizontal_irradiance.filter_by_hoys(hoys))

    def filter_by_moys(self, moys):
        """Create a new filtered Wea from this Wea based on a list of minutes of the year.

        Args:
           moys: A List of minutes of the year [0..8759 * 60].

        Returns:
            A new Wea with filtered data.
        """
        return Wea(
            self.location,
            self.direct_normal_irradiance.filter_by_moys(moys),
            self.diffuse_horizontal_irradiance.filter_by_moys(moys))

    def filter_by_sun_up(self, min_altitude=0):
        """Create a new filtered Wea from this Wea based on whether the sun is up

        Args:
            min_altitude: A number for the minimum altitude above the horizon at
                which the sun is considered up in degrees. Setting this to 0 will
                filter values for all hours where the sun is physically above the
                horizon. By setting this to a negative number (eg. -6), various levels
                of twilight can be used to filter the data (eg. civil twilight).
                Positive numbers can be used to discount low sun angles (Default: 0).

        Returns:
            A new Wea with filtered data.
        """
        sp = Sunpath.from_location(self.location)
        sp.is_leap_year = self.is_leap_year
        pattern = []
        for dt in self.datetimes:
            sun = sp.calculate_sun_from_date_time(dt)
            sun_up = True if sun.altitude > min_altitude else False
            pattern.append(sun_up)
        return self.filter_by_pattern(pattern)

    def get_irradiance_value(self, month, day, hour):
        """Get direct and diffuse irradiance values for a point in time.

        Args:
            month: Integer for month of the year [1 - 12].
            day: Integer for the day of the month [1 - 31].
            hour: Float for hour of the day [0 - 23].
        """
        dt = DateTime(month, day, hour, leap_year=self.is_leap_year)
        try:
            count = int(dt.hoy * self.timestep) if self.is_annual else \
                self.direct_normal_irradiance.datetimes.index(dt)
        except ValueError as e:
            raise ValueError('Datetime {} was not found in the Wea.\n{}'.format(dt, e))
        return self.direct_normal_irradiance[count], \
            self.diffuse_horizontal_irradiance[count]

    def get_irradiance_value_for_hoy(self, hoy):
        """Get direct and diffuse irradiance values for a hoy.

        Args:
            hoy: Float for hour of the year [0 - 8759].
        """
        try:
            count = int(hoy * self.timestep) if self.is_annual else \
                self.direct_normal_irradiance.datetimes.index(DateTime.from_hoy(hoy))
        except ValueError as e:
            raise ValueError('HOY {} was not found in the Wea.\n{}'.format(hoy, e))
        return self.direct_normal_irradiance[count], \
            self.diffuse_horizontal_irradiance[count]

    def directional_irradiance(self, altitude=90, azimuth=180,
                               ground_reflectance=0.2, isotropic=True):
        """Get the irradiance components for a surface facing a given direction.

        Note this method computes unobstructed solar flux facing a given
        altitude and azimuth. The default is set to return the global horizontal
        irradiance, assuming an altitude facing straight up (90 degrees).

        Args:
            altitude: A number between -90 and 90 that represents the
                altitude at which irradiance is being evaluated in degrees.
            azimuth: A number between 0 and 360 that represents the
                azimuth at which irradiance is being evaluated in degrees.
            ground_reflectance: A number between 0 and 1 that represents the
                reflectance of the ground. Default is set to 0.2. Some
                common ground reflectances are:

                *   urban: 0.18
                *   grass: 0.20
                *   fresh grass: 0.26
                *   soil: 0.17
                *   sand: 0.40
                *   snow: 0.65
                *   fresh_snow: 0.75
                *   asphalt: 0.12
                *   concrete: 0.30
                *   sea: 0.06

            isotropic: A boolean value that sets whether an isotropic sky is
                used (as opposed to an anisotropic sky). An isotropic sky
                assumes an even distribution of diffuse irradiance across the
                sky while an anisotropic sky places more diffuse irradiance
                near the solar disc. (Default: True).

        Returns:
            A tuple of four elements

            -   total_irradiance: A data collection of total solar irradiance.

            -   direct_irradiance: A data collection of direct solar irradiance.

            -   diffuse_irradiance: A data collection of diffuse sky solar irradiance.

            -   reflected_irradiance: A data collection of ground reflected solar
                irradiance.
        """
        # function to convert polar coordinates to xyz.
        def pol2cart(phi, theta):
            mult = math.cos(theta)
            x = math.sin(phi) * mult
            y = math.cos(phi) * mult
            z = math.sin(theta)
            return Vector3D(x, y, z)

        # convert the altitude and azimuth to a normal vector
        normal = pol2cart(math.radians(azimuth), math.radians(altitude))

        # create sunpath and get altitude at every timestep of the year
        dir_irr, diff_irr, ref_irr, total_irr = [], [], [], []
        sp = Sunpath.from_location(self.location)
        sp.is_leap_year = self.is_leap_year
        for dt, dnr, dhr in zip(self.datetimes, self.direct_normal_irradiance,
                                self.diffuse_horizontal_irradiance):
            sun = sp.calculate_sun_from_date_time(dt)
            sun_vec = pol2cart(math.radians(sun.azimuth),
                               math.radians(sun.altitude))
            vec_angle = sun_vec.angle(normal)

            # direct irradiance on surface
            srf_dir = 0
            if sun.altitude > 0 and vec_angle < math.pi / 2:
                srf_dir = dnr * math.cos(vec_angle)

            # diffuse irradiance on surface
            if isotropic:
                srf_dif = dhr * ((math.sin(math.radians(altitude)) / 2) + 0.5)
            else:
                y = max(0.45, 0.55 + (0.437 * math.cos(vec_angle)) + 0.313 *
                        math.cos(vec_angle) * 0.313 * math.cos(vec_angle))
                srf_dif = dhr * (y * (
                    math.sin(math.radians(abs(90 - altitude)))) +
                    math.cos(math.radians(abs(90 - altitude))))

            # reflected irradiance on surface.
            e_glob = dhr + dnr * math.cos(math.radians(90 - sun.altitude))
            srf_ref = e_glob * ground_reflectance * (0.5 - (math.sin(
                math.radians(altitude)) / 2))

            # add it all together
            dir_irr.append(srf_dir)
            diff_irr.append(srf_dif)
            ref_irr.append(srf_ref)
            total_irr.append(srf_dir + srf_dif + srf_ref)

        # create the headers
        data_head = Header(Irradiance(), 'W/m2', self.analysis_period, self.metadata)

        # create the data collections
        direct_irradiance = self._aligned_collection(data_head, dir_irr)
        diffuse_irradiance = self._aligned_collection(data_head, diff_irr)
        reflected_irradiance = self._aligned_collection(data_head, ref_irr)
        total_irradiance = self._aligned_collection(data_head, total_irr)

        return total_irradiance, direct_irradiance, \
            diffuse_irradiance, reflected_irradiance

    def estimate_illuminance_components(self, dew_point):
        """Get estimated direct, diffuse, and global illuminance from this Wea.

        Note that this method should only be used when there are no measured
        illuminance values that correspond to this Wea's irradiance values.
        Because the illuminance components calculated here are simply estimated
        using a model by Perez [1], they are not as accurate as true measured values.

        Note:
            [1] Perez R. (1990). 'Modeling Daylight Availability and Irradiance
            Components from Direct and Global Irradiance'. Solar Energy.
            Vol. 44. No. 5, pp. 271-289. USA.

        Args:
            dew_point: A data collection of dewpoint temperature in degrees C. This
                data collection must align with the irradiance data on this object.

        Returns:
            A tuple with four elements

            -   global_horiz_ill: Data collection of Global Horizontal Illuminance
                in lux.

            -   direct_normal_ill: Data collection of Direct Normal Illuminance in lux.

            -   diffuse_horizontal_ill: Data collection of  Diffuse Horizontal
                Illuminance in lux.

            -   zenith_lum: Data collection of Zenith Luminance in lux.
        """
        # check the dew_point input
        assert dew_point.is_collection_aligned(self.direct_normal_irradiance), \
            'Input dew_point data must be aligned with the irradiance on the Wea.'

        # calculate illuminance values
        sp = Sunpath.from_location(self.location)
        sp.is_leap_year = self.is_leap_year
        gh_ill_values, dn_ill_values, dh_ill_values, zen_lum_values = [], [], [], []
        for dt, dp, ghi, dni, dhi in zip(
                self.datetimes, dew_point, self.global_horizontal_irradiance,
                self.direct_normal_irradiance, self.diffuse_horizontal_irradiance):
            alt = sp.calculate_sun_from_date_time(dt).altitude
            gh, dn, dh, z = estimate_illuminance_from_irradiance(alt, ghi, dni, dhi, dp)
            gh_ill_values.append(gh)
            dn_ill_values.append(dn)
            dh_ill_values.append(dh)
            zen_lum_values.append(z)

        # create data collection headers for the results
        gh_ill_head = Header(GlobalHorizontalIlluminance(), 'lux',
                             self.analysis_period, self.metadata)
        dn_ill_head = Header(DirectNormalIlluminance(), 'lux',
                             self.analysis_period, self.metadata)
        dh_ill_head = Header(DiffuseHorizontalIlluminance(), 'lux',
                             self.analysis_period, self.metadata)
        zen_lum_head = Header(ZenithLuminance(), 'cd/m2',
                              self.analysis_period, self.metadata)

        # create data collections to hold illuminance results
        global_horiz_ill = self._aligned_collection(gh_ill_head, gh_ill_values)
        direct_normal_ill = self._aligned_collection(dn_ill_head, dn_ill_values)
        diffuse_horizontal_ill = self._aligned_collection(dh_ill_head, dh_ill_values)
        zenith_lum = self._aligned_collection(zen_lum_head, zen_lum_values)

        return global_horiz_ill, direct_normal_ill, diffuse_horizontal_ill, zenith_lum

    def to_dict(self):
        """Get the Wea as a dictionary."""
        base = {
            'type': 'Wea',
            'location': self.location.to_dict(),
            'direct_normal_irradiance': self.direct_normal_irradiance.values,
            'diffuse_horizontal_irradiance': self.diffuse_horizontal_irradiance.values,
            'timestep': self.timestep,
            'is_leap_year': self.is_leap_year
        }
        if not self.is_annual:
            dts = self.direct_normal_irradiance.datetimes
            base['datetimes'] = [dat.to_array() for dat in dts]
        return base

    def to_file_string(self):
        """Get a text string for the entirety of the Wea file contents."""
        lines = [self.header]
        for dir_rad, dif_rad, dt in zip(self.direct_normal_irradiance,
                                        self.diffuse_horizontal_irradiance,
                                        self.datetimes):
            line = "%d %d %.3f %d %d\n" \
                % (dt.month, dt.day, dt.float_hour, dir_rad, dif_rad)
            lines.append(line)
        return ''.join(lines)

    def write(self, file_path, write_hours=False):
        """Write the Wea object to a .wea file and return the file path.

        Args:
            file_path: Text string for the path to where the .wea file should be written.
            write_hours: Boolean to note whether a .hrs file should be written
                next to the .wea file, which lists the hours of the year (hoys)
                contained within the .wea file.
        """
        # write the .wea file
        if not file_path.lower().endswith('.wea'):
            file_path += '.wea'
        file_data = self.to_file_string()
        write_to_file(file_path, file_data, True)

        # write the .hrs file if requested
        if write_hours:
            hrs_file_path = file_path[:-4] + '.hrs'
            hrs_data = ','.join(str(h) for h in self.hoys) + '\n'
            write_to_file(hrs_file_path, hrs_data, True)
        return file_path

    def duplicate(self):
        """Duplicate location."""
        return self.__copy__()

    @staticmethod
    def to_constant_value(wea_file, value=1000):
        """Convert a Wea file to have a constant value for each datetime.

        This is useful in workflows where hourly irradiance values are inconsequential
        to the analysis and one is only using the Wea as a format to pass location
        and datetime information (eg. for direct sun hours).

        Args:
            wea_file: Full path to .wea file.
            value: The direct and diffuse irradiance value that will be written
                in for all datetimes of the Wea.

        Returns:
            Text string of Wea file contents with all irradiance values replaces
            with the input value.
        """
        assert os.path.isfile(wea_file), 'Failed to find {}'.format(wea_file)
        new_lines, value = [], str(int(value))
        with open(wea_file, readmode) as weaf:
            for i in range(6):
                new_lines.append(weaf.readline())
            for line in weaf:
                vals = line.split()
                vals[-2], vals[-1] = value, value
                new_lines.append(' '.join(vals) + '\n')
        return ''.join(new_lines)

    @staticmethod
    def count_timesteps(wea_file):
        """Count the number of timesteps represented within a Wea file.

        This is useful in workflows where one needs to compute cumulative values
        over a Wea (eg. cumulative radiation).

        Args:
            wea_file: Full path to .wea file.

        Returns:
            integer for the number of timesteps in the Wea.
        """
        assert os.path.isfile(wea_file), 'Failed to find {}'.format(wea_file)
        with open(wea_file, readmode) as weaf:
            count = len(weaf.readlines()) - 6
        return count

    def _aligned_collection(self, header, values):
        """Process a header and values into a collection aligned with Wea data."""
        if self.is_continuous:
            return HourlyContinuousCollection(header, values)
        else:
            dts = self.direct_normal_irradiance.datetimes
            return HourlyDiscontinuousCollection(header, values, dts)

    @staticmethod
    def _get_datetimes(timestep, is_leap_year):
        """Get a list of annual datetimes based on timestep.

        This method should only be used for classmethods. For datetimes use
        datetimes or hoys methods.
        """
        hour_count = 8760 + 24 if is_leap_year else 8760
        adjust_time = 30 if timestep == 1 else 0
        return tuple(
            DateTime.from_moy(60.0 * count / timestep + adjust_time, is_leap_year)
            for count in xrange(hour_count * timestep)
        )

    @staticmethod
    def _get_data_collections(dnr_values, dhr_values, metadata, timestep, is_leap_year):
        """Return two annual data collections for Direct Normal, Diffuse Horizontal."""
        analysis_period = AnalysisPeriod(timestep=timestep, is_leap_year=is_leap_year)
        dnr_header = Header(data_type=DirectNormalIrradiance(),
                            unit='W/m2',
                            analysis_period=analysis_period,
                            metadata=metadata)
        direct_norm_rad = HourlyContinuousCollection(dnr_header, dnr_values)
        dhr_header = Header(data_type=DiffuseHorizontalIrradiance(),
                            unit='W/m2',
                            analysis_period=analysis_period,
                            metadata=metadata)
        diffuse_horiz_rad = HourlyContinuousCollection(dhr_header, dhr_values)

        return direct_norm_rad, diffuse_horiz_rad

    @staticmethod
    def _parse_wea_header(weaf, wea_file_name):
        """Parse the Ladybug location from a wea header given the wea file object."""
        first_line = weaf.readline()
        assert first_line.startswith('place'), 'Failed to find place in .wea header.\n' \
            '{} is not a valid wea file.'.format(wea_file_name)
        location = Location()
        location.city = ' '.join(first_line.split()[1:])
        location.latitude = float(weaf.readline().split()[-1])
        location.longitude = -float(weaf.readline().split()[-1])
        location.time_zone = -int(weaf.readline().split()[-1]) / 15
        location.elevation = float(weaf.readline().split()[-1])
        weaf.readline()  # pass line for weather data units
        return location

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __len__(self):
        return len(self.direct_normal_irradiance)

    def __getitem__(self, key):
        return self.direct_normal_irradiance[key], \
            self.diffuse_horizontal_irradiance[key]

    def __iter__(self):
        return zip(self.direct_normal_irradiance.values,
                   self.diffuse_horizontal_irradiance.values)

    def __key(self):
        return self.location, self.direct_horizontal_irradiance, \
            self.diffuse_horizontal_irradiance

    def __eq__(self, other):
        return isinstance(other, Wea) and self.__key() == other.__key()

    def __ne__(self, value):
        return not self.__eq__(value)

    def __copy__(self):
        new_wea = Wea(
            self.location.duplicate(),
            self.direct_normal_irradiance.duplicate(),
            self.diffuse_horizontal_irradiance.duplicate()
        )
        new_wea._enforce_on_hour = self._enforce_on_hour
        new_wea.metadata = deepcopy(self.metadata)
        return new_wea

    def __repr__(self):
        """Wea object representation."""
        return "WEA [%s]" % self.location.city



# coding=utf-8
"""Object for calculating PMV comfort from DataCollections."""
from __future__ import division

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.polyline import Polyline2D
from ladybug_geometry.geometry2d.mesh import Mesh2D
from ladybug_geometry.geometry3d.pointvector import Point3D

from .epw import EPW
from .datacollection import DailyCollection, HourlyContinuousCollection, \
    HourlyDiscontinuousCollection
from .psychrometrics import humid_ratio_from_db_rh, db_temp_from_enth_hr, \
    db_temp_from_rh_hr, db_temp_and_hr_from_wb_rh
from .legend import LegendParameters
from .graphic import GraphicContainer

from .datatype.time import Time
from .datatype.temperature import Temperature
from .datatype.fraction import Fraction
from .datatype.specificenergy import Enthalpy


class PsychrometricChart(object):
    """Class for constructing psychrometric charts and plotting data on them.

    Args:
        temperature: Hourly, daily, or sub-hourly data collection of temperature
            values in Celsius or a single temperature value to be used for the
            whole analysis.
        relative_humidity: Hourly, daily, or sub-hourly data collection of relative
            humidity values in % or a single relative humidity value to be used
            for the whole analysis.
        average_pressure: Number for the average air pressure across the data
            plotted on the chart (Pa). (Default: 101325 Pa; pressure at sea level).
        legend_parameters: An optional LegendParameter object to change the display
            of the PsychrometricChart. (Default: None).
        base_point: A Point2D to be used as a starting point to generate the geometry
            of the plot. (Default: (0, 0)).
        x_dim: A number to set the X dimension of each degree of temperature on the
            chart. (Default: 1).
        y_dim: A number to set the Y dimension of a unity humidity ratio on the chart.
            Note that most maximum humidity ratios are around 0.03. (Default: 1500).
        min_temperature: An integer for the minimum temperature on the chart in
            degrees. This should be celsius if use_ip is False and fahrenheit if
            use_ip is True. (Default: -20; suitable for celsius).
        max_temperature: An integer for the maximum temperature on the chart in
            degrees. This should be celsius if use_ip is False and fahrenheit if
            use_ip is True. (Default: 50; suitable for celsius).
        max_humidity_ratio: A value for the maximum humidity ratio in kg water / kg
            air. (Default: 0.03).
        use_ip: Boolean to note whether temperature values should be plotted in
            Fahrenheit instead of Celsius. (Default: False).

    Properties:
        * temperature
        * relative_humidity
        * average_pressure
        * legend_parameters
        * base_point
        * x_dim
        * y_dim
        * min_temperature
        * max_temperature
        * max_humidity_ratio
        * use_ip
        * saturation_line
        * chart_border
        * temperature_labels
        * temperature_label_points
        * temperature_lines
        * rh_labels
        * rh_label_points
        * rh_lines
        * hr_labels
        * hr_label_points
        * hr_lines
        * enthalpy_labels
        * enthalpy_label_points
        * enthalpy_lines
        * wb_labels
        * wb_label_points
        * wb_lines
        * title_text
        * title_location
        * x_axis_text
        * x_axis_location
        * y_axis_text
        * y_axis_location
        * data_points
        * time_matrix
        * hour_values
        * colored_mesh
        * legend
        * container
    """
    ACCEPTABLE_COLLECTIONS = (DailyCollection, HourlyContinuousCollection,
                              HourlyDiscontinuousCollection)
    TEMP_TYPE = Temperature()
    ENTH_TYPE = Enthalpy()

    def __init__(self, temperature, relative_humidity, average_pressure=101325,
                 legend_parameters=None, base_point=Point2D(), x_dim=1, y_dim=1500,
                 min_temperature=-20, max_temperature=50, max_humidity_ratio=0.03,
                 use_ip=False): 
        """Initialize Psychrometric Chart."""
        # check and assign the temperature and humidity
        self._use_ip = bool(use_ip)
        self._calc_length = 1
        self._time_multiplier = 1
        self._temperature = temperature
        self._relative_humidity = relative_humidity
        self._t_values = self._t_values_c = self._check_input(
            temperature, Temperature, 'C', 'temperature')
        self._rh_values = self._check_input(
            relative_humidity, Fraction, '%', 'relative_humidity')
        if len(self._t_values) == 1:
            self._t_values = self._t_values_c = self._t_values * self._calc_length
        if self._use_ip:  # convert everything to Fahrenheit
            self._t_values = self.TEMP_TYPE.to_unit(self._t_values, 'F', 'C')
        assert len(self._t_values) == len(self._rh_values), \
            'Number of temperature and humidity values must match.'

        # assign the inputs as properties of the chart
        self._average_pressure = self._check_number(average_pressure, 'average_pressure')
        assert isinstance(base_point, Point2D), 'Expected Point2D for ' \
            'PsychrometricChart base point. Got {}.'.format(type(base_point))
        self._base_point = base_point
        self._x_dim = self._check_number(x_dim, 'x_dim')
        self._y_dim = self._check_number(y_dim, 'y_dim')
        assert max_temperature - min_temperature >= 10, 'Psychrometric chart ' \
            'max_temperature and min_temperature difference must be at least 10.'
        self._max_temperature = int(max_temperature)
        self._min_temperature = int(min_temperature)
        self._max_humidity_ratio = float(max_humidity_ratio)
        assert self._max_humidity_ratio >= 0.005, 'Psychrometric chart ' \
            'max_humidity_ratio must be at least 0.005.'

        # create the graphic container
        if self._use_ip:  # categorize based on every 1.66 fahrenheit
            self._t_category = []
            current_t, max_t = self._min_temperature, self._max_temperature + 1.75
            while current_t < max_t:
                current_t += (5 / 3)
                self._t_category.append(current_t)
        else:  # categorize based on every degree celsius
            self._t_category = list(range(self._min_temperature + 1,
                                          self._max_temperature + 1))
        self._rh_category = list(range(5, 105, 5))
        self._time_matrix, self._hour_values, self._remove_pattern = \
            self._compute_hour_values()
        assert len(self._hour_values) > 0, \
            'No data was found to lie on the psychrometric chart.'
        max_x = base_point.x + (self._max_temperature - self._min_temperature + 5) \
            * self._x_dim
        max_pt = Point3D(max_x, self.hr_y_value(self.max_humidity_ratio), 0)
        min_pt = Point3D(base_point.x, base_point.y, 0)
        self._container = GraphicContainer(
            self._hour_values, min_pt, max_pt, legend_parameters, Time(), 'hr')
        self._process_legend_default(self._container.legend_parameters)

        # create global attributes used by several of the geometry properties
        self._temp_range = list(range(self._min_temperature, self._max_temperature, 5)) \
            + [self._max_temperature]
        self._x_range = [self.t_x_value(t) for t in self._temp_range]
        if use_ip:  # ensure that _temp_range is always in celsius
            self._temp_range = self.TEMP_TYPE.to_unit(self._temp_range, 'C', 'F')
        rh_range = range(10, 110, 10)
        self._rh_lines = tuple(self.relative_humidity_polyline(rh) for rh in rh_range)
        self._saturation_line = self.relative_humidity_polyline(100, 2)
        max_hr_thnd = int(self._max_humidity_ratio * 1000)
        base_hr_range = list(range(5, max_hr_thnd, 5)) + [max_hr_thnd]
        max_db_hr = 1000 * humid_ratio_from_db_rh(
            self._temp_range[-1], 100, self._average_pressure)
        base_hr_range = [val for val in base_hr_range if val <= max_db_hr]
        self._hr_range = tuple(round(val / 1000, 3) for val in base_hr_range)
        self._y_range = [self._y_dim * hr + self._base_point.y for hr in self._hr_range]

        # set null values for properties that are optional
        self._chart_border = None
        self._enth_range = None
        self._enth_lines = None
        self._wb_range = None
        self._wb_lines = None
        self._data_points = None
        self._colored_mesh = None

        # check to be sure we don't have conditions above the boiling point
        assert self._temp_range[-1] < 100, \
            'Temperatures above the boiling point of water are not plot-able.'

    @classmethod
    def from_epw(cls, epw_file, legend_parameters=None, base_point=Point2D(),
                 x_dim=1, y_dim=1500, min_temperature=-20, max_temperature=50,
                 max_humidity_ratio=0.03, use_ip=False):
        """Create a psychrometric chart object using the data in an epw file.

        Args:
            epw_file: Full path to epw weather file.
            legend_parameters: An optional LegendParameter object to change the display
                of the PsychrometricChart. (Default: None).
            base_point: A Point2D to be used as a starting point to generate the geometry
                of the plot. (Default: (0, 0)).
            x_dim: A number to set the X dimension of each degree of temperature on the
                chart. (Default: 1).
            y_dim: A number to set the Y dimension of unity humidity ratio on the chart.
                Note that most maximum humidity ratios are around 0.03. (Default: 1500).
            min_temperature: An integer for the minimum temperature on the chart in
                degrees. This should be celsius if use_ip is False and fahrenheit if
                use_ip is True. (Default: -20; suitable for celsius).
            max_temperature: An integer for the maximum temperature on the chart in
                degrees. This should be celsius if use_ip is False and fahrenheit if
                use_ip is True. (Default: 50; suitable for celsius).
            max_humidity_ratio: A value for the maximum humidity ratio in kg water / kg
                air. (Default: 0.03).
            use_ip: Boolean to note whether temperature values should be plotted in
                Fahrenheit instead of Celsius. (Default: False).
        """
        epw = EPW(epw_file)
        pressure = epw.atmospheric_station_pressure.average
        return cls(
            epw.dry_bulb_temperature, epw.relative_humidity, pressure, legend_parameters,
            base_point, x_dim, y_dim, min_temperature, max_temperature,
            max_humidity_ratio, use_ip)

    @classmethod
    def from_dict(cls, data):
        """ Create PsychrometricChart from a dictionary

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            'type': 'PsychrometricChart',
            'temperature': {},  # data collection or value for temperature [C]
            'relative_humidity': {},  # data collection or value for humidity [%]
            'average_pressure': 101325,  # average atmospheric pressure [Pa]
            'legend_parameters': {},  # legend parameters dictionary
            'base_point': {},  # Point2D dictionary
            'x_dim': 1.0,  # value for X dimension per degree
            'y_dim': 1500.0,  # value for Y dimension for unity humidity ratio
            'min_temperature': -20.0,  # value for minimum temperature
            'max_temperature': 50.0,  # value for maximum temperature
            'max_humidity_ratio': 0.03,  # value for maximum humidity ratio
            'use_ip': False,  # boolean for whether to use IP values
            }
        """
        # process the optional inputs
        p = data['average_pressure'] if 'average_pressure' in data else 101325
        lp = LegendParameters.from_dict(data['legend_parameters']) \
            if 'legend_parameters' in data else None
        bpt = Point2D.from_dict(data['base_point']) if 'base_point' in data \
            else Point2D()
        xd = data['x_dim'] if 'x_dim' in data else 1
        yd = data['y_dim'] if 'y_dim' in data else 1500
        tmin = data['min_temperature'] if 'min_temperature' in data else -20
        tmax = data['max_temperature'] if 'max_temperature' in data else 50
        hrmax = data['max_humidity_ratio'] if 'max_humidity_ratio' in data else 0.03
        ip = data['use_ip'] if 'use_ip' in data else False

        # process the data collections
        class_mapper = {
            'DailyCollection': DailyCollection,
            'HourlyContinuousCollection': HourlyContinuousCollection,
            'HourlyDiscontinuousCollection': HourlyDiscontinuousCollection}
        t_data, rh_data = data['temperature'], data['relative_humidity']
        temp = class_mapper[t_data['type']].from_dict(t_data) \
            if isinstance(t_data, dict) else t_data
        rh = class_mapper[rh_data['type']].from_dict(rh_data) \
            if isinstance(rh_data, dict) else rh_data
        return cls(temp, rh, p, lp, bpt, xd, yd, tmin, tmax, hrmax, ip)

    @property
    def temperature(self):
        """The temperature assigned to this psychrometric chart [C]."""
        return self._temperature

    @property
    def relative_humidity(self):
        """The relative humidity assigned to this psychrometric chart."""
        return self._relative_humidity

    @property
    def average_pressure(self):
        """the average air pressure across the data plotted on the chart (Pa)."""
        return self._average_pressure

    @property
    def legend_parameters(self):
        """The legend parameters customizing this psychrometric chart."""
        return self._container.legend_parameters

    @property
    def base_point(self):
        """Point3D for the base point of this psychrometric chart."""
        return self._base_point

    @property
    def x_dim(self):
        """The X dimension of each degree of temperature on the chart."""
        return self._x_dim

    @property
    def y_dim(self):
        """The Y dimension of a unity humidity ratio on the chart."""
        return self._y_dim

    @property
    def min_temperature(self):
        """An integer for the minimum temperature on the chart.

        Will be in celsius if use_ip is False and fahrenheit if use_ip is True.
        """
        return self._min_temperature

    @property
    def max_temperature(self):
        """An integer for the maximum temperature on the chart.

        Will be in celsius if use_ip is False and fahrenheit if use_ip is True.
        """
        return self._max_temperature

    @property
    def max_humidity_ratio(self):
        """A value for the maximum humidity ratio in kg water / kg air."""
        return self._max_humidity_ratio

    @property
    def use_ip(self):
        """Boolean for whether temperature should be in Fahrenheit or Celsius."""
        return self._use_ip

    @property
    def saturation_line(self):
        """Get a Polyline2D for the saturation line of the chart."""
        return self._saturation_line

    @property
    def chart_border(self):
        """Get a Polyline2D for the border of the chart (excluding saturation line)."""
        if self._chart_border is None:
            self._chart_border = self._compute_border()
        return self._chart_border

    @property
    def temperature_labels(self):
        """Get a tuple of text for the temperature labels on the chart."""
        if self.use_ip:
            temp_range = tuple(range(self._min_temperature, self._max_temperature, 5)) \
                + (self._max_temperature,)
            return tuple(str(val) for val in temp_range)
        return tuple(str(val) for val in self._temp_range)

    @property
    def temperature_label_points(self):
        """Get a tuple of Point2Ds for the temperature labels on the chart."""
        y_val = self._base_point.y - self.legend_parameters.text_height * 0.5
        return tuple(Point2D(x_val, y_val) for x_val in self._x_range)

    @property
    def temperature_lines(self):
        """Get a tuple of LineSegment2Ds for the temperature labels on the chart."""
        # get the Y-values for the top of the temperature lines
        hr_vals = (humid_ratio_from_db_rh(t, 100, self.average_pressure)
                   for t in self._temp_range)
        top_y = []
        for hr in hr_vals:
            y_val = self.hr_y_value(hr) if hr < self._max_humidity_ratio \
                else self.hr_y_value(self._max_humidity_ratio)
            top_y.append(y_val)

        t_lines = []  # create the array of line segments
        for x_val, y_val in zip(self._x_range, top_y):
            l_seg = LineSegment2D.from_end_points(
                Point2D(x_val, self._base_point.y), Point2D(x_val, y_val))
            t_lines.append(l_seg)
        return t_lines

    @property
    def rh_labels(self):
        """Get a tuple of text for the relative humidity labels on the chart."""
        return tuple('{}%'.format(val) for val in range(10, 110, 10))

    @property
    def rh_label_points(self):
        """Get a tuple of Point2Ds for the relative humidity labels on the chart."""
        last_sgs = (LineSegment2D.from_end_points(p[-2], p[-1]) for p in self._rh_lines)
        last_dirs = (seg.v.reverse().normalize() * (self._x_dim * 2) for seg in last_sgs)
        move_vec = (Vector2D(vec.x - (self._x_dim * 0.4), vec.y) for vec in last_dirs)
        return tuple(pl[-1].move(vec) for pl, vec in zip(self._rh_lines, move_vec))

    @property
    def rh_lines(self):
        """Get a tuple of Polyline2Ds for the relative humidity labels on the chart."""
        return self._rh_lines

    @property
    def hr_labels(self):
        """Get a tuple of text for the humidity ratio labels on the chart."""
        return tuple(str(val) for val in self._hr_range)

    @property
    def hr_label_points(self):
        """Get a tuple of Point2Ds for the humidity ratio labels on the chart."""
        x_val = self._x_range[-1] + self.legend_parameters.text_height * 0.5
        return tuple(Point2D(x_val, y_val) for y_val in self._y_range)

    @property
    def hr_lines(self):
        """Get a tuple of LineSegment2Ds for the humidity ratio labels on the chart."""
        hr_lines, xmax = [], self._x_range[-1]
        for hr, y in zip(self._hr_range, self._y_range):
            tmin = db_temp_from_rh_hr(100, hr, self.average_pressure)
            tmin = self.TEMP_TYPE.to_unit([tmin], 'F', 'C')[0] if self.use_ip else tmin
            xmin = self.t_x_value(tmin)
            xmin = xmin if xmin > self.base_point.x else self.base_point.x
            l_seg = LineSegment2D.from_end_points(Point2D(xmax, y), Point2D(xmin, y))
            hr_lines.append(l_seg)
        return hr_lines

    @property
    def enthalpy_labels(self):
        """Get a tuple of text for the enthalpy labels on the chart."""
        if self._enth_range is None:
            self._compute_enthalpy_range()
        return tuple('{} kJ/kg'.format(val) for val in self._enth_range) if not \
            self.use_ip else tuple('{} Btu/lb'.format(val) for val in self._enth_range)

    @property
    def enthalpy_label_points(self):
        """Get a tuple of Point2Ds for the humidity ratio labels on the chart."""
        if self._enth_lines is None:
            self._compute_enthalpy_range()
        return self._labels_points_from_lines(self._enth_lines)

    @property
    def enthalpy_lines(self):
        """Get a tuple of LineSegment2Ds for the humidity ratio labels on the chart."""
        if self._enth_lines is None:
            self._compute_enthalpy_range()
        return self._enth_lines

    @property
    def wb_labels(self):
        """Get a tuple of text for the wet bulb labels on the chart."""
        if self._wb_range is None:
            self._compute_wb_range()
        return tuple('{} C'.format(val) for val in self._wb_range) if not \
            self.use_ip else tuple('{} F'.format(val) for val in self._wb_range)

    @property
    def wb_label_points(self):
        """Get a tuple of Point2Ds for the wet bulb labels on the chart."""
        if self._wb_lines is None:
            self._compute_wb_range()
        return self._labels_points_from_lines(self._wb_lines)

    @property
    def wb_lines(self):
        """Get a tuple of LineSegment2Ds for the wet bulb temp labels on the chart."""
        if self._wb_lines is None:
            self._compute_wb_range()
        return self._wb_lines

    @property
    def title_text(self):
        """Get text for the title of the chart."""
        title_items = ['Time [hr]']
        extra_data = []
        if isinstance(self.temperature, self.ACCEPTABLE_COLLECTIONS):
            extra_data = self.temperature.header.metadata.items()
        elif isinstance(self.relative_humidity, self.ACCEPTABLE_COLLECTIONS):
            extra_data = self.relative_humidity.header.metadata.items()
        return '\n'.join(title_items + ['{}: {}'.format(k, v) for k, v in extra_data])

    @property
    def title_location(self):
        """Get a Point2D for the title of the chart."""
        origin = self.container.upper_title_location.o
        return Point2D(origin.x, origin.y)

    @property
    def x_axis_text(self):
        """Get text for the X-axis label of the chart."""
        unit = 'C' if not self.use_ip else 'F'
        if isinstance(self.temperature, self.ACCEPTABLE_COLLECTIONS):
            if 'type' in self.temperature.header.metadata:
                return '{} [{}]'.format(self.temperature.header.metadata['type'], unit)
            else:
                return '{} [{}]'.format(self.temperature.header.data_type, unit)
        return 'Temperature [{}]'.format(unit)

    @property
    def x_axis_location(self):
        """Get a Point2D for the X-axis label of the chart."""
        y_val = self._base_point.y - self.legend_parameters.text_height * 2.5
        return Point2D(self.base_point.x, y_val)

    @property
    def y_axis_text(self):
        """Get text for the Y-axis label of the chart."""
        unit = 'kg' if not self.use_ip else 'lb'
        return 'Humidity Ratio\n[{0} water / {0} air]'.format(unit)

    @property
    def y_axis_location(self):
        """Get a Point2D for the Y-axis label of the chart."""
        x_val = self._container.max_point.x + self.legend_parameters.text_height * 1.5
        return Point2D(x_val, self._container.max_point.y)

    @property
    def data_points(self):
        """Get a tuple of Point2Ds for each of the temperature and humidity values."""
        if self._data_points is None:
            p = self._average_pressure
            self._data_points = tuple(
                Point2D(
                    self.t_x_value(t), self.hr_y_value(humid_ratio_from_db_rh(c, r, p)))
                for t, c, r in zip(self._t_values, self._t_values_c, self._rh_values))
        return self._data_points

    @property
    def time_matrix(self):
        """Get a tuple of of tuples where each sub-tuple is a row of the mesh.

        Each value in the resulting matrix corresponds to the number of temperature/
        humidity points in a given cell of the mesh.
        """
        return tuple(tuple(row) for row in self._time_matrix)

    @property
    def hour_values(self):
        """Get a tuple for the number of hours associated with each colored_mesh face."""
        return self._hour_values

    @property
    def colored_mesh(self):
        """Get a colored mesh for the number of hours for each part of the chart."""
        if self._colored_mesh is None:
            self._colored_mesh = self._generate_mesh()
        return self._colored_mesh

    @property
    def legend(self):
        """The legend assigned to this graphic."""
        return self._container._legend

    @property
    def container(self):
        """Get the GraphicContainer for the colored mesh."""
        return self._container

    def plot_point(self, temperature, relative_humidity):
        """Get a Point2D for a given temperature and relative humidity on the chart.

        Args:
            temperature: A temperature value, which should be in Celsius if use_ip
                is False and Fahrenheit is use_ip is True.
            relative_humidity: A relative humidity value in % (from 0 to 100).
        """
        tc = temperature if not self.use_ip else \
            self.TEMP_TYPE.to_unit([temperature], 'C', 'F')[0]
        hr = humid_ratio_from_db_rh(tc, relative_humidity, self.average_pressure)
        return Point2D(self.t_x_value(temperature), self.hr_y_value(hr))

    def data_mesh(self, data_collection, legend_parameters=None):
        """Get a colored mesh for a data_collection aligned with the chart's data.

        Args:
            data_collection: A data collection that is aligned with the temperature
                and humidity values of the chart.
            legend_parameters: Optional legend parameters to customize the legend
                and look of the resulting mesh.

        Returns:
            A tuple with two values.

            -   mesh: A colored Mesh2D similar to the chart's colored_mesh property
                but where each face is colored with the average value of the input
                data_collection.

            -   container: A GraphicContainer object for the mesh, which possesses
                a legend that corresponds to the mesh.
        """
        # check to be sure the data collection aligns
        data_vals = data_collection.values
        assert len(data_vals) == self._calc_length, 'Number of data collection values ' \
            'must match those of the psychometric chart temperature and humidity.'

        # create a matrix with a tally of the hours for all the data
        base_mtx = [[[] for val in self._t_category] for rh in self._rh_category]
        for t, rh, val in zip(self._t_values, self._rh_values, data_vals):
            if t < self._min_temperature or t > self._max_temperature:
                continue  # temperature value does not currently fit on the chart
            for y, rh_cat in enumerate(self._rh_category):
                if rh < rh_cat:
                    break
            for x, t_cat in enumerate(self._t_category):
                if t < t_cat:
                    break
            base_mtx[y][x].append(val)

        # compute average values
        avg_values = [sum(val_list) / len(val_list) for rh_l in base_mtx
                      for val_list in rh_l if len(val_list) != 0]

        # create the colored mesh and graphic container
        base_contain = self.container
        container = GraphicContainer(
            avg_values, base_contain.min_point, base_contain.max_point,
            legend_parameters, data_collection.header.data_type,
            data_collection.header.unit)
        self._process_legend_default(container.legend_parameters)
        mesh = self.colored_mesh.duplicate()  # start with hour mesh as a base
        mesh.colors = container.value_colors
        return mesh, container

    def relative_humidity_polyline(self, rh, subdivisions=1):
        """Get a Polyline2D for a given relative humidity value.

        Args:
            rh: A number between 0 and 100 for the relative humidity line to draw.
            subdivisions: Integer for the number of subdivisions for every 5
                degrees. (Default: 1).
        """
        # get the HR values and temperatures
        prs = self.average_pressure
        if subdivisions == 1:
            hr_vals = [humid_ratio_from_db_rh(t, rh, prs) for t in self._temp_range]
            x_vals = self._x_range
        else:  # build up custom temperatures and HRs
            hr_vals = [humid_ratio_from_db_rh(self._temp_range[0], rh, prs)]
            x_vals = [self._x_range[0]]
            t_diff = (self._temp_range[1] - self._temp_range[0]) / subdivisions
            x_diff = (self._x_range[1] - self._x_range[0]) / subdivisions
            for i in range(len(self._temp_range) - 1):
                st_t, st_x = self._temp_range[i], self._x_range[i]
                for j in range(subdivisions):
                    t = st_t + (j + 1) * t_diff
                    hr_vals.append(humid_ratio_from_db_rh(t, rh, prs))
                    x_vals.append(st_x + (j + 1) * x_diff)

        # loop through the values and create the points
        pts = []
        for i, (x, hr) in enumerate(zip(x_vals, hr_vals)):
            if hr < self._max_humidity_ratio:
                pts.append(Point2D(x, self.hr_y_value(hr)))
            else:  # we're at the top of the chart; cut it off
                if abs(self._max_humidity_ratio - hr_vals[i - 1]) < 0.001:
                    del pts[-1]  # avoid the case of a bad interpolation
                last_db = db_temp_from_rh_hr(
                    rh, self._max_humidity_ratio, self.average_pressure)
                last_db = self.TEMP_TYPE.to_unit([last_db], 'F', 'C')[0] \
                    if self.use_ip else last_db
                x_val = self.t_x_value(last_db)
                pts.append(Point2D(x_val, self.hr_y_value(self._max_humidity_ratio)))
                break
        return Polyline2D(pts, interpolated=True)

    def hr_y_value(self, humidity_ratio):
        """Get the Y-coordinate associated with a certain HR on the chart.

        Args:
            humidity_ratio: A humidity ratio value in kg water / kg air.
        """
        return self.base_point.y + humidity_ratio * self._y_dim

    def t_x_value(self, temperature):
        """Get the X-coordinate associated with a certain temperature on the chart.

        Args:
            temperature: A temperature value, which should be in Celsius if use_ip
                is False and Fahrenheit is use_ip is True.
        """
        return self._base_point.x + self._x_dim * (temperature - self._min_temperature)

    def to_dict(self):
        """Get psychrometric chart as a dictionary."""
        temp = self.temperature
        temp = temp.to_dict() if isinstance(temp, self.ACCEPTABLE_COLLECTIONS) else temp
        rh = self.relative_humidity
        rh = rh.to_dict() if isinstance(rh, self.ACCEPTABLE_COLLECTIONS) else rh
        return {
            'temperature': temp,
            'relative_humidity': rh,
            'average_pressure': self.average_pressure,
            'legend_parameters': self.legend_parameters.to_dict(),
            'base_point': self.base_point.to_dict(),
            'x_dim': self.x_dim,
            'y_dim': self.y_dim,
            'min_temperature': self.min_temperature,
            'max_temperature': self.max_temperature,
            'max_humidity_ratio': self.max_humidity_ratio,
            'use_ip': self.use_ip,
            'type': 'PsychrometricChart'
        }

    def _compute_hour_values(self):
        """Compute the matrix of binned time values based on the chart inputs.

        Returns:
            A tuple with three values.

            -   base_mtx: A full matrix with counts of values for each degree
                temperature and 5% RH of the chart.

            -   mesh_values: A list of numbers for the values of the mesh.

            -   remove_pattern: A list of booleans for which faces of the full mesh
                should be removed.
        """
        # create a matrix with a tally of the hours for all the data
        base_mtx = [[0 for val in self._t_category] for rh in self._rh_category]
        for t, rh in zip(self._t_values, self._rh_values):
            if t < self._min_temperature or t > self._max_temperature:
                continue  # temperature value does not currently fit on the chart
            for y, rh_cat in enumerate(self._rh_category):
                if rh < rh_cat:
                    break
            for x, t_cat in enumerate(self._t_category):
                if t < t_cat:
                    break
            base_mtx[y][x] += 1

        # flatten the matrix and create a pattern to remove faces
        flat_values = [tc * self._time_multiplier for rh_l in base_mtx for tc in rh_l]
        remove_pattern = [val != 0 for val in flat_values]
        mesh_values = tuple(val for val in flat_values if val != 0)
        return base_mtx, mesh_values, remove_pattern

    def _generate_mesh(self):
        """Get the colored mesh from this object's hour values."""
        # global properties used in the generation of the mesh
        prs = self.average_pressure
        t_per_row = [self._min_temperature] + self._t_category
        x_per_row = [self.t_x_value(t) for t in t_per_row]
        temp_in_c = self.TEMP_TYPE.to_unit(t_per_row, 'C', 'F') \
            if self.use_ip else t_per_row

        # loop through RH rows and create mesh vertices and faces
        vertices = [Point2D(x, self._base_point.y) for x in x_per_row]
        faces, vert_count, row_len = [], 0, len(t_per_row)
        for rh in self._rh_category:
            vert_count += row_len
            y1 = self.hr_y_value(humid_ratio_from_db_rh(temp_in_c[0], rh, prs))
            vertices.append(Point2D(x_per_row[0], y1))
            for i, t in enumerate(temp_in_c[1:]):
                y = self.hr_y_value(humid_ratio_from_db_rh(t, rh, prs))
                vertices.append(Point2D(x_per_row[i + 1], y))
                v1 = vert_count - row_len + i
                v2 = v1 + 1
                v3 = vert_count + i + 1
                v4 = v3 - 1
                faces.append((v1, v2, v3, v4))

        # create the Mesh2D, remove unused faces, and assign the colors
        mesh = Mesh2D(vertices, faces)
        mesh = mesh.remove_faces_only(self._remove_pattern)
        mesh.colors = self._container.value_colors
        return mesh

    def _compute_border(self):
        """Compute a Polyline2D for the outer border of the chart."""
        # get properties used to establish the border of the chart
        prs, bpt, hmax = self.average_pressure, self.base_point, self.max_humidity_ratio
        max_hr = humid_ratio_from_db_rh(self._temp_range[-1], 100, prs)
        y_left = self.hr_y_value(humid_ratio_from_db_rh(self._temp_range[0], 100, prs))
        y_right = self.hr_y_value(hmax) if max_hr > hmax else self.hr_y_value(max_hr)
        x_max = bpt.x + (self.max_temperature - self.min_temperature) * self._x_dim

        # get the points and build the polyline
        pt1, pt2, pt3, pt4 = \
            Point2D(bpt.x, y_left), bpt, Point2D(x_max, bpt.y), Point2D(x_max, y_right)
        if max_hr > hmax:
            return Polyline2D((pt1, pt2, pt3, pt4, self._saturation_line[-1]))
        return Polyline2D((pt1, pt2, pt3, pt4))

    def _compute_enthalpy_range(self):
        """Compute the values for enthalpy range and lines."""
        # constants used throughout the calculation
        low_y = self.base_point.y + 1e-6
        up_y = self.hr_y_value(self._max_humidity_ratio)
        border, sat_line = self.chart_border, self._saturation_line
        all_enthalpies, ref_temp = tuple(range(0, 160, 10)), 0
        enth_lbl = all_enthalpies
        if self.use_ip:
            enth_lbl = tuple(range(0, 65, 5))
            all_enthalpies = self.ENTH_TYPE.to_unit(enth_lbl, 'kJ/kg', 'Btu/lb')
            ref_temp = self.TEMP_TYPE.to_unit([0], 'C', 'F')[0]

        # loop through the enthalpies and compute the lines of constant enthalpy
        enth_range, enth_lines = [], []
        for i, enthalpy in enumerate(all_enthalpies):
            st_db = db_temp_from_enth_hr(enthalpy, 0.0, ref_temp)
            end_db = db_temp_from_enth_hr(enthalpy, 0.03, ref_temp)
            if self.use_ip:
                st_db, end_db = self.TEMP_TYPE.to_unit((st_db, end_db), 'F', 'C')
            enth_line = LineSegment2D.from_end_points(
                Point2D(self.t_x_value(st_db), low_y),
                Point2D(self.t_x_value(end_db), up_y))
            border_ints = border.intersect_line_ray(enth_line)
            if len(border_ints) == 2:
                enth_range.append(enth_lbl[i])
                seg = LineSegment2D.from_end_points(border_ints[0], border_ints[1])
                enth_lines.append(seg)
            else:
                sat_ints = sat_line.intersect_line_ray(enth_line)
                if len(sat_ints) != 0:
                    enth_range.append(enth_lbl[i])
                    if len(border_ints) == 1:
                        seg = LineSegment2D.from_end_points(border_ints[0], sat_ints[0])
                    else:
                        seg = LineSegment2D.from_end_points(enth_line.p, sat_ints[0])
                    enth_lines.append(seg)

        # set the properties on this class
        self._enth_range = enth_range
        self._enth_lines = enth_lines

    def _compute_wb_range(self):
        """Compute the values for wet bulb range and lines."""
        # constants used throughout the calculation
        low_y, border = self.base_point.y - 1e-6, self.chart_border
        all_wbs = wb_c = tuple(range(self._min_temperature, self._max_temperature, 5))
        if self.use_ip:
            wb_c = self.TEMP_TYPE.to_unit(wb_c, 'C', 'F')

        # loop through the wet bulb and compute the lines of constant wet bulb
        wb_range, wb_lines = [], []
        for i, wb in enumerate(wb_c):
            st_db = db_temp_and_hr_from_wb_rh(wb, 0, self._average_pressure)[0]
            end_db, end_hr = db_temp_and_hr_from_wb_rh(wb, 100, self._average_pressure)
            if self.use_ip:
                st_db, end_db = self.TEMP_TYPE.to_unit((st_db, end_db), 'F', 'C')
            enth_line = LineSegment2D.from_end_points(
                Point2D(self.t_x_value(st_db), low_y),
                Point2D(self.t_x_value(end_db), self.hr_y_value(end_hr)))
            border_ints = border.intersect_line_ray(enth_line)
            if len(border_ints) == 2:
                wb_range.append(all_wbs[i])
                seg = LineSegment2D.from_end_points(border_ints[0], border_ints[1])
                wb_lines.append(seg)
            elif len(border_ints) == 1:
                wb_range.append(all_wbs[i])
                seg = LineSegment2D.from_end_points(border_ints[0], enth_line.p2)
                wb_lines.append(seg)

        # set the properties on this class
        self._wb_range = wb_range
        self._wb_lines = wb_lines

    def _labels_points_from_lines(self, label_lines):
        """Extract label points from lines."""
        move_vec = []
        base_pts = []
        for seg in label_lines:
            if seg.v.y < 0:
                move_vec.append(seg.v.reverse().normalize() * self._x_dim * 1.5)
                base_pts.append(seg.p1)
            else:
                move_vec.append(seg.v.normalize() * self._x_dim * 1.5)
                base_pts.append(seg.p2)
        return tuple(pt.move(vec) for pt, vec in zip(base_pts, move_vec))

    def _process_legend_default(self, l_par):
        """Override the dimensions of the legend to ensure it fits the chart."""
        min_pt, max_pt = self.container.min_point, self.container.max_point
        if l_par.vertical and l_par.is_segment_height_default:
            l_par.properties_3d.segment_height = (max_pt.y - min_pt.y) / 20
            l_par.properties_3d._is_segment_height_default = True
        elif l_par.vertical and l_par.is_segment_height_default:
            l_par.properties_3d.segment_width = (max_pt.x - min_pt.x) / 20
            l_par.properties_3d._is_segment_width_default = True

    def _check_input(self, data_coll, dat_type, unit, name):
        """Check an input that can be either a number or a Data Collection."""
        if isinstance(data_coll, self.ACCEPTABLE_COLLECTIONS):
            self._check_datacoll(data_coll, dat_type, unit, name)
            return data_coll.values
        else:
            try:  # assume that it's a single number
                value = float(data_coll)
                return [value] * self._calc_length
            except (ValueError, TypeError):
                raise TypeError('{} must be either a number or a hourly/daily data '
                                'collection. Got {}'.format(name, type(data_coll)))

    def _check_datacoll(self, data_coll, dat_type, unit, name):
        """Check the data type and units of a Data Collection."""
        assert isinstance(data_coll.header.data_type, dat_type) and \
            data_coll.header.unit == unit, '{} must be {} in {}. ' \
            'Got {} in {}'.format(name, dat_type().name, unit,
                                  data_coll.header.data_type.name,
                                  data_coll.header.unit)
        if isinstance(data_coll, DailyCollection):
            self._time_multiplier = 24
        else:  # it's an hourly or sub-hourly collection
            self._time_multiplier = 1 / data_coll.header.analysis_period.timestep
        self._calc_length = len(data_coll)

    @staticmethod
    def _check_number(value, value_name):
        """Check a given value for a dimension input."""
        try:
            value = float(value)
        except (ValueError, TypeError):
            raise TypeError('Expected number for Psychrometric Chart {}. '
                            'Got {}.'.format(value_name, type(value)))
        assert value > 0, 'Psychrometric Chart {} must be greater than 0. ' \
            'Got {}.'.format(value_name, value)
        return value

    def __len__(self):
        """Return length of values on the object."""
        return len(self._t_values)

    def __getitem__(self, key):
        """Return a tuple of temperature and humidity."""
        return self._t_values[key], self._rh_values[key]

    def __iter__(self):
        """Iterate through the values."""
        return zip(self._t_values, self._rh_values)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Psychrometric Chart representation."""
        return 'Psychrometric Chart: {} values'.format(len(self._t_values))



# coding=utf-8
"""Pressure data type."""
from __future__ import division

from .base import DataTypeBase


class Pressure(DataTypeBase):
    """Pressure
    """
    _units = ('Pa', 'inHg', 'atm', 'bar', 'Torr', 'psi', 'inH2O')
    _si_units = ('Pa', 'bar')
    _ip_units = ('inHg', 'psi', 'inH2O')
    _abbreviation = 'P'
    _point_in_time = False

    def _Pa_to_inHg(self, value):
        return value * 0.0002953

    def _Pa_to_atm(self, value):
        return value / 101325.

    def _Pa_to_bar(self, value):
        return value / 100000.

    def _Pa_to_Torr(self, value):
        return value * 0.00750062

    def _Pa_to_psi(self, value):
        return value * 0.000145038

    def _Pa_to_inH2O(self, value):
        return value * 0.00401865

    def _inHg_to_Pa(self, value):
        return value / 0.0002953

    def _atm_to_Pa(self, value):
        return value * 101325.

    def _bar_to_Pa(self, value):
        return value * 100000.

    def _Torr_to_Pa(self, value):
        return value / 0.00750062

    def _psi_to_Pa(self, value):
        return value / 0.000145038

    def _inH2O_to_Pa(self, value):
        return value / 0.00401865

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('Pa', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'inHg', from_unit), 'inHg'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'Pa', from_unit), 'Pa'


class AtmosphericStationPressure(Pressure):
    _min = 0
    _abbreviation = 'Patm'
